"""
Functions to print a Python file with the definition of a dialect, given specifications
of its operations and attributes.
"""

import itertools
import subprocess
from collections.abc import Iterable
from typing import TextIO

from xdsl.ir import Attribute, ParametrizedAttribute, TypeAttribute
from xdsl.irdl import (
    AnyInt,
    OpDef,
    OperandDef,
    OptOperandDef,
    OptResultDef,
    ParamAttrDef,
    RangeOf,
    ResultDef,
    SingleOf,
    VarOperandDef,
    VarResultDef,
)


def generate_dynamic_attr_class(
    class_name: str, attr: ParamAttrDef, is_type: bool = True
) -> type[Attribute]:
    """
    Dynamically define a type based on ParamAttrDef.
    This is needed to reference dynamically created attributes in operations.
    """
    return type(
        class_name,
        (ParametrizedAttribute,) + ((TypeAttribute,) if is_type else ()),
        dict(ParametrizedAttribute.__dict__) | {"name": attr.name},
    )


def get_str_from_operand_or_result(
    name: str, operand_or_result: OperandDef | ResultDef
) -> str:
    """
    Get a constraint from the GenericRangeConstraint wrapper.
    Build the correct definition function based on the wrapper's type.
    """
    match operand_or_result.constr:
        case SingleOf():
            inner_constr = operand_or_result.constr.constr
        case RangeOf(length=AnyInt()):
            inner_constr = operand_or_result.constr.constr
        case _:
            raise NotImplementedError(
                f"Constraint type {operand_or_result.constr} not yet implemented"
            )

    match operand_or_result:
        case VarOperandDef():
            def_str = "var_operand_def"
        case OptOperandDef():
            def_str = "opt_operand_def"
        case OperandDef():
            def_str = "operand_def"
        case VarResultDef():
            def_str = "var_result_def"
        case OptResultDef():
            def_str = "opt_result_def"
        case ResultDef():
            def_str = "result_def"

    return f"{name} = {def_str}({inner_constr})"


def typedef_to_class_string(class_name: str, typedef: ParamAttrDef) -> str:
    """
    Generate class definition for a type.
    """
    if typedef.parameters:
        raise NotImplementedError("Attribute parameters not yet implemented")

    return f"""
@irdl_attr_definition
class {class_name}(ParametrizedAttribute, TypeAttribute):
\tname = "{typedef.name}"
    """


def attrdef_to_class_string(class_name: str, attr: ParamAttrDef) -> str:
    """
    Generate class definition for an attribute.
    """
    if attr.parameters:
        raise NotImplementedError("Attribute parameters not yet implemented")
    return f"""
@irdl_attr_definition
class {class_name}(ParametrizedAttribute):
\tname = "{attr.name}"
    """


def opdef_to_class_string(class_name: str, op: OpDef) -> str:
    """
    Generate class definition for an operation.
    """
    if op.accessor_names:
        raise NotImplementedError("Operation accessor_names not yet implemented")

    fields_description = ""

    fields_description += (
        "\n\t".join(
            [
                get_str_from_operand_or_result(name, operand_or_result)
                for name, operand_or_result in itertools.chain(op.operands, op.results)
            ]
        )
        + "\n\t"
    )

    if op.attributes:
        raise NotImplementedError("Operation attributes not yet implemented")
    if op.regions:
        raise NotImplementedError("Operation regions not yet implemented")
    if op.successors:
        raise NotImplementedError("Operation successors not yet implemented")
    if op.traits.traits:
        raise NotImplementedError(f"Operation traits not yet implemented {op.traits}")
    if op.properties:
        raise NotImplementedError("Operation properties not yet implemented")

    return f"""
@irdl_op_definition
class {class_name}(IRDLOperation):
\tname = "{op.name}"
\t{fields_description}
\t{f'assembly_format = "{op.assembly_format}"' if op.assembly_format else ""}
    """


def dump_dialect_pyfile(
    dialect_name: str,
    ops: Iterable[tuple[str, OpDef]] = (),
    *,
    attributes: Iterable[tuple[str, ParamAttrDef]] = (),
    types: Iterable[tuple[str, ParamAttrDef]] = (),
    out: TextIO | None = None,
    dialect_obj_name: str = "",
):
    """
    Generate a python file with a dialect comprised of given ops, attributes and types.
    """
    if not dialect_obj_name:
        dialect_obj_name = dialect_name.capitalize() + "Dialect"

    imports = """
from xdsl.dialects.builtin import *
from xdsl.ir import *
from xdsl.irdl import *

# ruff: noqa: F403, F405
    """

    types_class_defs = "\n".join(
        typedef_to_class_string(class_name, attr) for class_name, attr in types
    )

    attr_class_defs = "\n".join(
        attrdef_to_class_string(class_name, attr) for class_name, attr in attributes
    )

    op_class_defs = "\n".join(
        opdef_to_class_string(class_name, op) for class_name, op in ops
    )

    op_list = ",".join(name for name, _ in ops)
    attr_list = ",".join(name for name, _ in itertools.chain(attributes, types))

    dialect_def = (
        f'{dialect_obj_name} = Dialect("{dialect_name}", [{op_list}], [{attr_list}])'
    )

    content = "\n".join(
        (
            imports,
            types_class_defs,
            attr_class_defs,
            op_class_defs,
            dialect_def,
        )
    )
    # Format output
    output = subprocess.run(
        [
            "ruff",
            "format",
            "--stdin-filename",
            f"{dialect_name}.py",
        ],
        input=content,
        capture_output=True,
        text=True,
    )

    print(output.stdout, file=out, end="")



from collections.abc import Callable

from xdsl.passes import ModulePass


def get_all_passes() -> dict[str, Callable[[], type[ModulePass]]]:
    """Return the list of all available passes."""

    def get_apply_individual_rewrite():
        from xdsl.transforms.individual_rewrite import ApplyIndividualRewritePass

        return ApplyIndividualRewritePass

    def get_apply_pdl():
        from xdsl.transforms import apply_pdl

        return apply_pdl.ApplyPDLPass

    def get_arith_add_fastmath():
        from xdsl.transforms import arith_add_fastmath

        return arith_add_fastmath.AddArithFastMathFlagsPass

    def get_canonicalize_dmp():
        from xdsl.transforms import canonicalize_dmp

        return canonicalize_dmp.CanonicalizeDmpPass

    def get_canonicalize():
        from xdsl.transforms import canonicalize

        return canonicalize.CanonicalizePass

    def get_constant_fold_interp():
        from xdsl.transforms import constant_fold_interp

        return constant_fold_interp.ConstantFoldInterpPass

    def get_control_flow_hoist():
        from xdsl.transforms import control_flow_hoist

        return control_flow_hoist.ControlFlowHoistPass

    def get_convert_arith_to_riscv_snitch():
        from xdsl.backend.riscv.lowering import convert_arith_to_riscv_snitch

        return convert_arith_to_riscv_snitch.ConvertArithToRiscvSnitchPass

    def get_convert_arith_to_riscv():
        from xdsl.backend.riscv.lowering import convert_arith_to_riscv

        return convert_arith_to_riscv.ConvertArithToRiscvPass

    def get_convert_arith_to_varith():
        from xdsl.transforms import varith_transformations

        return varith_transformations.ConvertArithToVarithPass

    def get_convert_func_to_riscv_func():
        from xdsl.backend.riscv.lowering import convert_func_to_riscv_func

        return convert_func_to_riscv_func.ConvertFuncToRiscvFuncPass

    def get_convert_func_to_x86_func():
        from xdsl.backend.x86.lowering import convert_func_to_x86_func

        return convert_func_to_x86_func.ConvertFuncToX86FuncPass

    def get_convert_linalg_to_loops():
        from xdsl.transforms import convert_linalg_to_loops

        return convert_linalg_to_loops.ConvertLinalgToLoopsPass

    def get_convert_linalg_to_memref_stream():
        from xdsl.transforms import convert_linalg_to_memref_stream

        return convert_linalg_to_memref_stream.ConvertLinalgToMemRefStreamPass

    def get_convert_memref_stream_to_loops():
        from xdsl.transforms import convert_memref_stream_to_loops

        return convert_memref_stream_to_loops.ConvertMemRefStreamToLoopsPass

    def get_convert_memref_stream_to_snitch_stream():
        from xdsl.transforms import convert_memref_stream_to_snitch_stream

        return (
            convert_memref_stream_to_snitch_stream.ConvertMemRefStreamToSnitchStreamPass
        )

    def get_convert_memref_to_ptr():
        from xdsl.transforms import convert_memref_to_ptr

        return convert_memref_to_ptr.ConvertMemRefToPtr

    def get_convert_memref_to_riscv():
        from xdsl.backend.riscv.lowering import convert_memref_to_riscv

        return convert_memref_to_riscv.ConvertMemRefToRiscvPass

    def get_convert_ml_program_to_memref():
        from xdsl.transforms import convert_ml_program_to_memref

        return convert_ml_program_to_memref.ConvertMlProgramToMemRefPass

    def get_convert_print_format_to_riscv_debug():
        from xdsl.backend.riscv.lowering import convert_print_format_to_riscv_debug

        return convert_print_format_to_riscv_debug.ConvertPrintFormatToRiscvDebugPass

    def get_convert_ptr_to_riscv():
        from xdsl.transforms import convert_ptr_to_riscv

        return convert_ptr_to_riscv.ConvertPtrToRiscvPass

    def get_convert_ptr_type_offsets():
        from xdsl.transforms import convert_ptr_type_offsets

        return convert_ptr_type_offsets.ConvertPtrTypeOffsetsPass

    def get_convert_riscv_scf_for_to_frep():
        from xdsl.transforms import convert_riscv_scf_for_to_frep

        return convert_riscv_scf_for_to_frep.ConvertRiscvScfForToFrepPass

    def get_convert_riscv_scf_to_riscv_cf():
        from xdsl.backend.riscv.lowering import convert_riscv_scf_to_riscv_cf

        return convert_riscv_scf_to_riscv_cf.ConvertRiscvScfToRiscvCfPass

    def get_convert_riscv_to_llvm():
        from xdsl.transforms import convert_riscv_to_llvm

        return convert_riscv_to_llvm.ConvertRiscvToLLVMPass

    def get_convert_scf_to_cf():
        from xdsl.transforms import convert_scf_to_cf

        return convert_scf_to_cf.ConvertScfToCf

    def get_convert_scf_to_openmp():
        from xdsl.transforms import convert_scf_to_openmp

        return convert_scf_to_openmp.ConvertScfToOpenMPPass

    def get_convert_scf_to_riscv_scf():
        from xdsl.backend.riscv.lowering import convert_scf_to_riscv_scf

        return convert_scf_to_riscv_scf.ConvertScfToRiscvPass

    def get_convert_snitch_stream_to_snitch():
        from xdsl.backend.riscv.lowering import convert_snitch_stream_to_snitch

        return convert_snitch_stream_to_snitch.ConvertSnitchStreamToSnitch

    def get_convert_stencil_to_csl_stencil():
        from xdsl.transforms import convert_stencil_to_csl_stencil

        return convert_stencil_to_csl_stencil.ConvertStencilToCslStencilPass

    def get_convert_stencil_to_ll_mlir():
        from xdsl.transforms.experimental import convert_stencil_to_ll_mlir

        return convert_stencil_to_ll_mlir.ConvertStencilToLLMLIRPass

    def get_convert_varith_to_arith():
        from xdsl.transforms import varith_transformations

        return varith_transformations.ConvertVarithToArithPass

    def get_jax_use_donated_arguments():
        from xdsl.transforms import jax_use_donated_arguments

        return jax_use_donated_arguments.JaxUseDonatedArguments

    def get_cse():
        from xdsl.transforms import common_subexpression_elimination

        return common_subexpression_elimination.CommonSubexpressionElimination

    def get_csl_stencil_bufferize():
        from xdsl.transforms import csl_stencil_bufferize

        return csl_stencil_bufferize.CslStencilBufferize

    def get_csl_stencil_handle_async_flow():
        from xdsl.transforms import csl_stencil_handle_async_flow

        return csl_stencil_handle_async_flow.CslStencilHandleAsyncControlFlow

    def get_csl_stencil_materialize_stores():
        from xdsl.transforms import csl_stencil_materialize_stores

        return csl_stencil_materialize_stores.CslStencilMaterializeStores

    def get_csl_stencil_set_global_coeffs():
        from xdsl.transforms import csl_stencil_set_global_coeffs

        return csl_stencil_set_global_coeffs.CslStencilSetGlobalCoeffs

    def get_csl_stencil_to_csl_wrapper():
        from xdsl.transforms import csl_stencil_to_csl_wrapper

        return csl_stencil_to_csl_wrapper.CslStencilToCslWrapperPass

    def get_csl_wrapper_hoist_buffers():
        from xdsl.transforms import csl_wrapper_hoist_buffers

        return csl_wrapper_hoist_buffers.CslWrapperHoistBuffers

    def get_dce():
        from xdsl.transforms import dead_code_elimination

        return dead_code_elimination.DeadCodeElimination

    def get_distribute_stencil():
        from xdsl.transforms.experimental.dmp import stencil_global_to_local

        return stencil_global_to_local.DistributeStencilPass

    def get_dmp_to_mpi():
        from xdsl.transforms.experimental.dmp import stencil_global_to_local

        return stencil_global_to_local.DmpToMpiPass

    def get_empty_tensor_to_alloc_tensor():
        from xdsl.transforms import empty_tensor_to_alloc_tensor

        return empty_tensor_to_alloc_tensor.EmptyTensorToAllocTensorPass

    def get_eqsat_add_costs():
        from xdsl.transforms import eqsat_add_costs

        return eqsat_add_costs.EqsatAddCostsPass

    def get_eqsat_create_eclasses():
        from xdsl.transforms import eqsat_create_eclasses

        return eqsat_create_eclasses.EqsatCreateEclassesPass

    def get_eqsat_extract():
        from xdsl.transforms import eqsat_extract

        return eqsat_extract.EqsatExtractPass

    def get_frontend_desymrefy():
        from xdsl.frontend.pyast.passes.desymref import FrontendDesymrefyPass

        return FrontendDesymrefyPass

    def get_function_constant_pinning():
        from xdsl.transforms.experimental import function_constant_pinning

        return function_constant_pinning.FunctionConstantPinningPass

    def get_function_persist_arg_names():
        from xdsl.transforms import function_transformations

        return function_transformations.FunctionPersistArgNamesPass

    def get_gpu_map_parallel_loops():
        from xdsl.transforms import gpu_map_parallel_loops

        return gpu_map_parallel_loops.GpuMapParallelLoopsPass

    def get_hls_convert_stencil_to_ll_mlir():
        from xdsl.transforms.experimental import hls_convert_stencil_to_ll_mlir

        return hls_convert_stencil_to_ll_mlir.HLSConvertStencilToLLMLIRPass

    def get_inline_snrt():
        from xdsl.transforms import inline_snrt

        return inline_snrt.InlineSnrtPass

    def get_lift_arith_to_linalg():
        from xdsl.transforms.lift_arith_to_linalg import LiftArithToLinalg

        return LiftArithToLinalg

    def get_linalg_fuse_multiply_add():
        from xdsl.transforms.linalg_transformations import LinalgFuseMultiplyAddPass

        return LinalgFuseMultiplyAddPass

    def get_linalg_to_csl():
        from xdsl.transforms.linalg_to_csl import LinalgToCsl

        return LinalgToCsl

    def get_loop_hoist_memref():
        from xdsl.transforms import loop_hoist_memref

        return loop_hoist_memref.LoopHoistMemRefPass

    def get_lower_affine():
        from xdsl.transforms import lower_affine

        return lower_affine.LowerAffinePass

    def get_lower_csl_stencil():
        from xdsl.transforms import lower_csl_stencil

        return lower_csl_stencil.LowerCslStencil

    def get_lower_csl_wrapper():
        from xdsl.transforms import lower_csl_wrapper

        return lower_csl_wrapper.LowerCslWrapperPass

    def get_lower_hls():
        from xdsl.transforms.experimental import lower_hls

        return lower_hls.LowerHLSPass

    def get_lower_mpi():
        from xdsl.transforms import lower_mpi

        return lower_mpi.LowerMPIPass

    def get_lower_riscv_func():
        from xdsl.transforms import lower_riscv_func

        return lower_riscv_func.LowerRISCVFunc

    def get_lower_riscv_scf_to_labels():
        from xdsl.backend.riscv import riscv_scf_to_asm

        return riscv_scf_to_asm.LowerRiscvScfForToLabelsPass

    def get_lower_snitch():
        from xdsl.transforms import lower_snitch

        return lower_snitch.LowerSnitchPass

    def get_memref_stream_fold_fill():
        from xdsl.transforms import memref_stream_fold_fill

        return memref_stream_fold_fill.MemRefStreamFoldFillPass

    def get_memref_stream_generalize_fill():
        from xdsl.transforms import memref_stream_generalize_fill

        return memref_stream_generalize_fill.MemRefStreamGeneralizeFillPass

    def get_memref_stream_infer_fill():
        from xdsl.transforms import memref_stream_infer_fill

        return memref_stream_infer_fill.MemRefStreamInferFillPass

    def get_memref_stream_interleave():
        from xdsl.transforms import memref_stream_interleave

        return memref_stream_interleave.MemRefStreamInterleavePass

    def get_memref_stream_legalize():
        from xdsl.transforms import memref_stream_legalize

        return memref_stream_legalize.MemRefStreamLegalizePass

    def get_memref_stream_tile_outer_loops():
        from xdsl.transforms import memref_stream_tile_outer_loops

        return memref_stream_tile_outer_loops.MemRefStreamTileOuterLoopsPass

    def get_memref_stream_unnest_out_parameters():
        from xdsl.transforms import memref_stream_unnest_out_parameters

        return memref_stream_unnest_out_parameters.MemRefStreamUnnestOutParametersPass

    def get_memref_streamify():
        from xdsl.transforms import memref_streamify

        return memref_streamify.MemRefStreamifyPass

    def get_memref_to_dsd():
        from xdsl.transforms import memref_to_dsd

        return memref_to_dsd.MemRefToDsdPass

    def get_memref_to_gpu():
        from xdsl.transforms import gpu_allocs

        return gpu_allocs.MemRefToGPUPass

    def get_mlir_opt():
        from xdsl.transforms import mlir_opt

        return mlir_opt.MLIROptPass

    def get_printf_to_llvm():
        from xdsl.transforms import printf_to_llvm

        return printf_to_llvm.PrintfToLLVM

    def get_printf_to_putchar():
        from xdsl.transforms import printf_to_putchar

        return printf_to_putchar.PrintfToPutcharPass

    def get_reconcile_unrealized_casts():
        from xdsl.transforms import reconcile_unrealized_casts

        return reconcile_unrealized_casts.ReconcileUnrealizedCastsPass

    def get_replace_incompatible_fpga():
        from xdsl.transforms.experimental import replace_incompatible_fpga

        return replace_incompatible_fpga.ReplaceIncompatibleFPGA

    def get_riscv_register_allocation():
        from xdsl.transforms import riscv_register_allocation

        return riscv_register_allocation.RISCVRegisterAllocation

    def get_riscv_prologue_epilogue_insertion():
        from xdsl.backend.riscv import prologue_epilogue_insertion

        return prologue_epilogue_insertion.PrologueEpilogueInsertion

    def get_riscv_scf_loop_range_folding():
        from xdsl.transforms import riscv_scf_loop_range_folding

        return riscv_scf_loop_range_folding.RiscvScfLoopRangeFoldingPass

    def get_scf_for_loop_flatten():
        from xdsl.transforms import scf_for_loop_flatten

        return scf_for_loop_flatten.ScfForLoopFlattenPass

    def get_scf_for_loop_range_folding():
        from xdsl.transforms import scf_for_loop_range_folding

        return scf_for_loop_range_folding.ScfForLoopRangeFoldingPass

    def get_scf_parallel_loop_tiling():
        from xdsl.transforms import scf_parallel_loop_tiling

        return scf_parallel_loop_tiling.ScfParallelLoopTilingPass

    def get_shape_inference():
        from xdsl.transforms.shape_inference import ShapeInferencePass

        return ShapeInferencePass

    def get_snitch_allocate_registers():
        from xdsl.transforms import snitch_register_allocation

        return snitch_register_allocation.SnitchRegisterAllocation

    def get_stencil_bufferize():
        from xdsl.transforms import stencil_bufferize

        return stencil_bufferize.StencilBufferize

    def get_stencil_inlining():
        from xdsl.transforms import stencil_inlining

        return stencil_inlining.StencilInliningPass

    def get_stencil_shape_minimize():
        from xdsl.transforms import stencil_shape_minimize

        return stencil_shape_minimize.StencilShapeMinimize

    def get_stencil_storage_materialization():
        from xdsl.transforms.experimental import stencil_storage_materialization

        return stencil_storage_materialization.StencilStorageMaterializationPass

    def get_stencil_tensorize_z_dimension():
        from xdsl.transforms.experimental import stencil_tensorize_z_dimension

        return stencil_tensorize_z_dimension.StencilTensorizeZDimension

    def get_stencil_unroll():
        from xdsl.transforms import stencil_unroll

        return stencil_unroll.StencilUnrollPass

    def get_test_add_timers_to_top_level_funcs():
        from xdsl.transforms import function_transformations

        return function_transformations.TestAddBenchTimersToTopLevelFunctions

    def get_test_lower_linalg_to_snitch():
        from xdsl.transforms import test_lower_linalg_to_snitch

        return test_lower_linalg_to_snitch.TestLowerLinalgToSnitchPass

    def get_varith_fuse_repeated_operands():
        from xdsl.transforms import varith_transformations

        return varith_transformations.VarithFuseRepeatedOperandsPass

    # Please insert pass and `get_` function in alphabetical order

    return {
        "apply-individual-rewrite": get_apply_individual_rewrite,
        "apply-pdl": get_apply_pdl,
        "arith-add-fastmath": get_arith_add_fastmath,
        "canonicalize-dmp": get_canonicalize_dmp,
        "canonicalize": get_canonicalize,
        "constant-fold-interp": get_constant_fold_interp,
        "control-flow-hoist": get_control_flow_hoist,
        "convert-arith-to-riscv-snitch": get_convert_arith_to_riscv_snitch,
        "convert-arith-to-riscv": get_convert_arith_to_riscv,
        "convert-arith-to-varith": get_convert_arith_to_varith,
        "convert-func-to-riscv-func": get_convert_func_to_riscv_func,
        "convert-func-to-x86-func": get_convert_func_to_x86_func,
        "convert-linalg-to-loops": get_convert_linalg_to_loops,
        "convert-linalg-to-memref-stream": get_convert_linalg_to_memref_stream,
        "convert-memref-stream-to-loops": get_convert_memref_stream_to_loops,
        "convert-memref-stream-to-snitch-stream": get_convert_memref_stream_to_snitch_stream,
        "convert-memref-to-ptr": get_convert_memref_to_ptr,
        "convert-memref-to-riscv": get_convert_memref_to_riscv,
        "convert-ml-program-to-memref": get_convert_ml_program_to_memref,
        "convert-print-format-to-riscv-debug": get_convert_print_format_to_riscv_debug,
        "convert-ptr-to-riscv": get_convert_ptr_to_riscv,
        "convert-riscv-scf-for-to-frep": get_convert_riscv_scf_for_to_frep,
        "convert-riscv-scf-to-riscv-cf": get_convert_riscv_scf_to_riscv_cf,
        "convert-riscv-to-llvm": get_convert_riscv_to_llvm,
        "convert-scf-to-cf": get_convert_scf_to_cf,
        "convert-scf-to-openmp": get_convert_scf_to_openmp,
        "convert-scf-to-riscv-scf": get_convert_scf_to_riscv_scf,
        "convert-snitch-stream-to-snitch": get_convert_snitch_stream_to_snitch,
        "convert-ptr-type-offsets": get_convert_ptr_type_offsets,
        "convert-stencil-to-csl-stencil": get_convert_stencil_to_csl_stencil,
        "convert-stencil-to-ll-mlir": get_convert_stencil_to_ll_mlir,
        "convert-varith-to-arith": get_convert_varith_to_arith,
        "jax-use-donated-arguments": get_jax_use_donated_arguments,
        "cse": get_cse,
        "csl-stencil-bufferize": get_csl_stencil_bufferize,
        "csl-stencil-handle-async-flow": get_csl_stencil_handle_async_flow,
        "csl-stencil-materialize-stores": get_csl_stencil_materialize_stores,
        "csl-stencil-set-global-coeffs": get_csl_stencil_set_global_coeffs,
        "csl-stencil-to-csl-wrapper": get_csl_stencil_to_csl_wrapper,
        "csl-wrapper-hoist-buffers": get_csl_wrapper_hoist_buffers,
        "dce": get_dce,
        "distribute-stencil": get_distribute_stencil,
        "dmp-to-mpi": get_dmp_to_mpi,
        "empty-tensor-to-alloc-tensor": get_empty_tensor_to_alloc_tensor,
        "eqsat-add-costs": get_eqsat_add_costs,
        "eqsat-create-eclasses": get_eqsat_create_eclasses,
        "eqsat-extract": get_eqsat_extract,
        "frontend-desymrefy": get_frontend_desymrefy,
        "function-constant-pinning": get_function_constant_pinning,
        "function-persist-arg-names": get_function_persist_arg_names,
        "gpu-map-parallel-loops": get_gpu_map_parallel_loops,
        "hls-convert-stencil-to-ll-mlir": get_hls_convert_stencil_to_ll_mlir,
        "inline-snrt": get_inline_snrt,
        "lift-arith-to-linalg": get_lift_arith_to_linalg,
        "linalg-fuse-multiply-add": get_linalg_fuse_multiply_add,
        "linalg-to-csl": get_linalg_to_csl,
        "loop-hoist-memref": get_loop_hoist_memref,
        "lower-affine": get_lower_affine,
        "lower-csl-stencil": get_lower_csl_stencil,
        "lower-csl-wrapper": get_lower_csl_wrapper,
        "lower-hls": get_lower_hls,
        "lower-mpi": get_lower_mpi,
        "lower-riscv-func": get_lower_riscv_func,
        "lower-riscv-scf-to-labels": get_lower_riscv_scf_to_labels,
        "lower-snitch": get_lower_snitch,
        "memref-stream-fold-fill": get_memref_stream_fold_fill,
        "memref-stream-generalize-fill": get_memref_stream_generalize_fill,
        "memref-stream-infer-fill": get_memref_stream_infer_fill,
        "memref-stream-interleave": get_memref_stream_interleave,
        "memref-stream-legalize": get_memref_stream_legalize,
        "memref-stream-tile-outer-loops": get_memref_stream_tile_outer_loops,
        "memref-stream-unnest-out-parameters": get_memref_stream_unnest_out_parameters,
        "memref-streamify": get_memref_streamify,
        "memref-to-dsd": get_memref_to_dsd,
        "memref-to-gpu": get_memref_to_gpu,
        "mlir-opt": get_mlir_opt,
        "printf-to-llvm": get_printf_to_llvm,
        "printf-to-putchar": get_printf_to_putchar,
        "reconcile-unrealized-casts": get_reconcile_unrealized_casts,
        "replace-incompatible-fpga": get_replace_incompatible_fpga,
        "riscv-allocate-registers": get_riscv_register_allocation,
        "riscv-prologue-epilogue-insertion": get_riscv_prologue_epilogue_insertion,
        "riscv-scf-loop-range-folding": get_riscv_scf_loop_range_folding,
        "scf-for-loop-flatten": get_scf_for_loop_flatten,
        "scf-for-loop-range-folding": get_scf_for_loop_range_folding,
        "scf-parallel-loop-tiling": get_scf_parallel_loop_tiling,
        "shape-inference": get_shape_inference,
        "snitch-allocate-registers": get_snitch_allocate_registers,
        "stencil-bufferize": get_stencil_bufferize,
        "stencil-inlining": get_stencil_inlining,
        "stencil-shape-minimize": get_stencil_shape_minimize,
        "stencil-storage-materialization": get_stencil_storage_materialization,
        "stencil-tensorize-z-dimension": get_stencil_tensorize_z_dimension,
        "stencil-unroll": get_stencil_unroll,
        "test-add-timers-to-top-level-funcs": get_test_add_timers_to_top_level_funcs,
        "test-lower-linalg-to-snitch": get_test_lower_linalg_to_snitch,
        "varith-fuse-repeated-operands": get_varith_fuse_repeated_operands,
    }



import pytest

import xdsl.dialects.pdl as pdl
from xdsl.builder import Builder
from xdsl.dialects.builtin import ArrayAttr, IntegerAttr, StringAttr, i32, i64
from xdsl.ir import Block
from xdsl.irdl import IRDLOperation, irdl_op_definition, traits_def
from xdsl.traits import HasParent, IsTerminator
from xdsl.utils.exceptions import VerifyException
from xdsl.utils.test_value import TestSSAValue

type_type = pdl.TypeType()
attribute_type = pdl.AttributeType()
value_type = pdl.ValueType()
operation_type = pdl.OperationType()

block = Block(
    arg_types=[
        type_type,
        attribute_type,
        value_type,
        operation_type,
    ]
)

type_val, attr_val, val_val, op_val = block.args


def test_build_anc():
    anc = pdl.ApplyNativeConstraintOp("anc", [type_val])

    assert anc.constraint_name == StringAttr("anc")
    assert anc.args == (type_val,)


def test_build_anr():
    anr = pdl.ApplyNativeRewriteOp("anr", [type_val], [attribute_type])

    assert anr.constraint_name == StringAttr("anr")
    assert anr.args == (type_val,)
    assert len(anr.results) == 1
    assert anr.result_types == (attribute_type,)


def test_build_rewrite():
    r = pdl.RewriteOp(
        name="r", root=None, external_args=[type_val, attr_val], body=None
    )

    assert r.name_ == StringAttr("r")
    assert r.external_args == (type_val, attr_val)
    assert len(r.results) == 0
    assert r.body is None

    r1 = pdl.RewriteOp(name="r", root=None, external_args=[type_val, attr_val])

    assert r1.body is not None


def test_build_operation_replace():
    operation = pdl.OperationOp(
        op_name="operation",
        attribute_value_names=ArrayAttr([StringAttr("name")]),
        operand_values=[val_val],
        attribute_values=[attr_val],
        type_values=[type_val],
    )

    assert operation.opName == StringAttr("operation")
    assert operation.attributeValueNames == ArrayAttr([StringAttr("name")])
    assert operation.operand_values == (val_val,)
    assert operation.attribute_values == (attr_val,)
    assert operation.type_values == (type_val,)

    replace = pdl.ReplaceOp(op_value=op_val, repl_operation=operation.results[0])
    replace.verify()

    assert replace.op_value == op_val
    assert replace.repl_operation == operation.results[0]
    assert replace.repl_values == ()

    replace = pdl.ReplaceOp(op_value=op_val, repl_values=[val_val])
    replace.verify()

    assert replace.op_value == op_val
    assert replace.repl_operation is None
    assert replace.repl_values == (val_val,)

    with pytest.raises(VerifyException):
        replace = pdl.ReplaceOp(op_value=op_val)
        replace.verify()

    with pytest.raises(VerifyException):
        replace = pdl.ReplaceOp(
            op_value=op_val, repl_operation=operation.results[0], repl_values=[val_val]
        )
        replace.verify()


def test_build_pattern():
    @Builder.implicit_region
    def body() -> None:
        root = pdl.OperationOp("test.op")
        pdl.RewriteOp(root.results[0])

    pattern = pdl.PatternOp(1, "pattern", body)

    assert pattern.benefit == IntegerAttr.from_int_and_width(1, 16)
    assert pattern.sym_name == StringAttr("pattern")
    assert pattern.body is body

    @irdl_op_definition
    class DummyTerminatorOp(IRDLOperation):
        name = "dummy.terminator"
        traits = traits_def(HasParent(pdl.PatternOp), IsTerminator())

    with pytest.raises(
        VerifyException, match="expected body to terminate with a `pdl.rewrite`"
    ):

        @Builder.implicit_region
        def body() -> None:
            DummyTerminatorOp()

        pattern = pdl.PatternOp(1, "pattern", body)
        pattern.verify()

    with pytest.raises(
        VerifyException, match="the pattern must contain at least one `pdl.operation`"
    ):
        root = pdl.OperationOp("test.op")

        @Builder.implicit_region
        def body() -> None:
            pdl.RewriteOp(root.results[0])

        pattern = pdl.PatternOp(1, "pattern", body)
        pattern.verify()


def test_build_result():
    res = pdl.ResultOp(IntegerAttr.from_int_and_width(1, 32), parent=op_val)

    assert res.index == IntegerAttr.from_int_and_width(1, 32)
    assert res.parent_ == op_val


def test_build_resultS():
    res = pdl.ResultsOp(op_val)

    assert res.parent_ == op_val
    assert res.index is None
    assert res.val.type == pdl.RangeType(pdl.ValueType())


def test_build_results_with_index():
    res = pdl.ResultsOp(op_val, IntegerAttr.from_int_and_width(1, 32))

    assert res.parent_ == op_val
    assert res.index == IntegerAttr.from_int_and_width(1, 32)
    assert res.val.type == pdl.RangeType(pdl.ValueType())


def test_build_results_with_index_and_type():
    res = pdl.ResultsOp(op_val, IntegerAttr.from_int_and_width(1, 32), pdl.ValueType())

    assert res.parent_ == op_val
    assert res.index == IntegerAttr.from_int_and_width(1, 32)
    assert res.val.type == pdl.ValueType()


def test_build_type():
    pdl_type = pdl.TypeOp()
    assert pdl_type.constantType is None

    pdl_type = pdl.TypeOp(i32)
    assert pdl_type.constantType == i32


def test_build_types():
    pdl_type = pdl.TypesOp()
    assert pdl_type.constantTypes is None

    pdl_type = pdl.TypesOp((i32, i64, i32))
    assert pdl_type.constantTypes == ArrayAttr((i32, i64, i32))


def test_build_operand():
    operand = pdl.OperandOp(val_val)

    assert operand.value_type == val_val


def test_range():
    val1 = TestSSAValue(pdl.ValueType())
    val2 = TestSSAValue(pdl.RangeType(pdl.ValueType()))
    val3 = TestSSAValue(pdl.ValueType())

    range_op = pdl.RangeOp((val1, val2, val3))

    assert range_op.arguments == (val1, val2, val3)
    assert range_op.result.type == pdl.RangeType(pdl.ValueType())


def test_empty_range():
    return_type = pdl.RangeType(pdl.ValueType())
    empty_range = pdl.RangeOp((), return_type)

    assert len(empty_range.arguments) == 0
    assert empty_range.result.type == return_type


def test_range_cannot_infer():
    with pytest.raises(ValueError):
        pdl.RangeOp(())  # Cannot infer return type



from collections.abc import Iterator
from dataclasses import dataclass
from typing import Any

from xdsl.dialects import snitch_stream
from xdsl.interpreter import (
    Interpreter,
    InterpreterFunctions,
    PythonValues,
    impl,
    register_impls,
)
from xdsl.interpreters.utils import ptr
from xdsl.interpreters.utils.stream import ReadableStream, WritableStream


@dataclass
class StridedPointerInputStream(ReadableStream[float]):
    offset_iter: Iterator[int]
    pointer: ptr.RawPtr
    index = -1

    def read(self) -> float:
        self.index += 1
        offset = next(self.offset_iter)
        return ptr.TypedPtr((self.pointer + offset), xtype=ptr.float64)[0]


@dataclass
class StridedPointerOutputStream(WritableStream[float]):
    index = -1
    offset_iter: Iterator[int]
    pointer: ptr.RawPtr

    def write(self, value: float) -> None:
        self.index += 1
        offset = next(self.offset_iter)
        ptr.TypedPtr((self.pointer + offset), xtype=ptr.float64)[0] = value


@register_impls
class SnitchStreamFunctions(InterpreterFunctions):
    @impl(snitch_stream.StreamingRegionOp)
    def run_streaming_region(
        self,
        interpreter: Interpreter,
        op: snitch_stream.StreamingRegionOp,
        args: tuple[Any, ...],
    ) -> PythonValues:
        input_stream_count = len(op.inputs)
        output_stream_count = len(op.outputs)
        input_pointers: tuple[ptr.RawPtr, ...] = args[:input_stream_count]
        output_pointers: tuple[ptr.RawPtr, ...] = args[
            input_stream_count : input_stream_count + output_stream_count
        ]

        if len(op.stride_patterns) == 1:
            pattern = op.stride_patterns.data[0]
            input_stride_patterns = (pattern,) * input_stream_count
            output_stride_patterns = (pattern,) * output_stream_count
        else:
            input_stride_patterns = op.stride_patterns.data[:input_stream_count]
            output_stride_patterns = op.stride_patterns.data[input_stream_count:]

        input_streams = tuple(
            StridedPointerInputStream(pat.offset_iter(), ptr)
            for pat, ptr in zip(input_stride_patterns, input_pointers, strict=True)
        )

        output_streams = tuple(
            StridedPointerOutputStream(pat.offset_iter(), ptr)
            for pat, ptr in zip(output_stride_patterns, output_pointers, strict=True)
        )

        interpreter.run_ssacfg_region(
            op.body, (*input_streams, *output_streams), "steraming_region"
        )

        return ()



"""
An embedding of equivalence classes in IR, for use in equality saturation with
non-destructive rewrites.

Please see the Equality Saturation Project for details:
https://github.com/orgs/xdslproject/projects/23

TODO: add documentation once we have end-to-end flow working:
https://github.com/xdslproject/xdsl/issues/3174
"""

from __future__ import annotations

from typing import ClassVar

from xdsl.dialects.builtin import IntAttr
from xdsl.ir import Attribute, Dialect, SSAValue
from xdsl.irdl import (
    AnyAttr,
    IRDLOperation,
    VarConstraint,
    irdl_op_definition,
    opt_attr_def,
    result_def,
    var_operand_def,
)
from xdsl.utils.exceptions import DiagnosticException, VerifyException

EQSAT_COST_LABEL = "eqsat_cost"
"""
Key used to store the cost of computing the result of an operation.
"""


@irdl_op_definition
class EClassOp(IRDLOperation):
    T: ClassVar = VarConstraint("T", AnyAttr())

    name = "eqsat.eclass"
    arguments = var_operand_def(T)
    result = result_def(T)
    min_cost_index = opt_attr_def(IntAttr)

    assembly_format = "$arguments attr-dict `:` type($result)"

    def __init__(
        self,
        *arguments: SSAValue,
        min_cost_index: IntAttr | None = None,
        res_type: Attribute | None = None,
    ):
        if not arguments:
            raise DiagnosticException("eclass op must have at least one operand")
        if res_type is None:
            res_type = arguments[0].type

        super().__init__(
            operands=[arguments],
            result_types=[res_type],
            attributes={"min_cost_index": min_cost_index},
        )

    def verify_(self) -> None:
        # Check that none of the operands are produced by another eclass op.
        # In that case the two ops should have been merged into one.
        for operand in self.operands:
            if isinstance(operand.owner, EClassOp):
                raise VerifyException(
                    "A result of an eclass operation cannot be used as an operand of "
                    "another eclass."
                )


EqSat = Dialect(
    "eqsat",
    [
        EClassOp,
    ],
)



from dataclasses import dataclass
from math import prod
from typing import Any, cast

from xdsl.context import Context
from xdsl.dialects import arith, riscv, riscv_snitch
from xdsl.dialects.builtin import (
    Float16Type,
    Float32Type,
    Float64Type,
    ModuleOp,
    UnrealizedConversionCastOp,
    VectorType,
)
from xdsl.ir import Operation
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
)


@dataclass
class LowerBinaryFloatVectorOp(RewritePattern):
    arith_op_cls: type[arith.FloatingPointLikeBinaryOperation]
    riscv_d_op_cls: type[riscv.RdRsRsFloatOperationWithFastMath]
    riscv_snitch_v_f32_op_cls: type[riscv.RdRsRsFloatOperationWithFastMath]
    riscv_snitch_v_f16_op_cls: type[riscv.RdRsRsFloatOperationWithFastMath]

    def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter) -> None:
        if not isinstance(op, self.arith_op_cls):
            return

        operand_type = op.result.type
        if not isinstance(operand_type, VectorType):
            return
        shape = operand_type.shape
        count = prod(dim.data for dim in shape.data)

        operand_type = cast(VectorType[Any], operand_type)
        scalar_type = operand_type.element_type

        lhs = UnrealizedConversionCastOp.get(
            (op.lhs,), (riscv.Registers.UNALLOCATED_FLOAT,)
        )
        rhs = UnrealizedConversionCastOp.get(
            (op.rhs,), (riscv.Registers.UNALLOCATED_FLOAT,)
        )

        match scalar_type:
            case Float64Type():
                if count != 1:
                    return
                cls = self.riscv_d_op_cls
            case Float32Type():
                if count != 2:
                    return
                cls = self.riscv_snitch_v_f32_op_cls
            case Float16Type():
                if count != 4:
                    return
                cls = self.riscv_snitch_v_f16_op_cls
            case _:
                raise ValueError(f"Unexpected float type {op.lhs.type}")

        rv_flags = riscv.FastMathFlagsAttr("none")
        if op.fastmath is not None:
            rv_flags = riscv.FastMathFlagsAttr(op.fastmath.data)

        new_op = cls(lhs, rhs, rd=riscv.Registers.UNALLOCATED_FLOAT, fastmath=rv_flags)
        cast_op = UnrealizedConversionCastOp.get((new_op.rd,), (op.result.type,))

        rewriter.replace_matched_op((lhs, rhs, new_op, cast_op))


lower_arith_addf = LowerBinaryFloatVectorOp(
    arith.AddfOp, riscv.FAddDOp, riscv_snitch.VFAddSOp, riscv_snitch.VFAddHOp
)


class ConvertArithToRiscvSnitchPass(ModulePass):
    name = "convert-arith-to-riscv-snitch"

    def apply(self, ctx: Context, op: ModuleOp) -> None:
        walker = PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    lower_arith_addf,
                ]
            ),
            apply_recursively=False,
        )
        walker.rewrite_module(op)



import pytest

from xdsl.builder import Builder, ImplicitBuilder
from xdsl.dialects import func, riscv, riscv_scf
from xdsl.dialects.builtin import IndexType, ModuleOp
from xdsl.interpreter import Interpreter
from xdsl.interpreters.func import FuncFunctions
from xdsl.interpreters.riscv import RiscvFunctions
from xdsl.interpreters.riscv_scf import RiscvScfFunctions
from xdsl.ir import BlockArgument

index = IndexType()
register = riscv.Registers.UNALLOCATED_INT


def sum_to_for_fn(n: int) -> int:
    """
    Python implementation of sum_to_for_op
    """
    result = 0
    for i in range(0, n, 1):
        result += i
    return result


@ModuleOp
@Builder.implicit_region
def sum_to_for_op():
    with ImplicitBuilder(func.FuncOp("sum_to", ((register,), (register,))).body) as (
        ub,
    ):
        lb = riscv.LiOp(0)
        step = riscv.LiOp(1)
        initial = riscv.LiOp(0)

        @Builder.implicit_region((register, register))
        def for_loop_region(args: tuple[BlockArgument, ...]):
            (i, acc) = args
            res = riscv.AddOp(i, acc)
            riscv_scf.YieldOp(res)

        result = riscv_scf.ForOp(lb, ub, step, (initial,), for_loop_region)
        func.ReturnOp(result)


# Python implementation of `sum_to_while_op`
# def sum_to_while_fn(n: int) -> int:
#     result = 0
#     i = 0
#     while i < n:
#         result += 1
#         i += 1
#     return result


@ModuleOp
@Builder.implicit_region
def sum_to_while_op():
    with ImplicitBuilder(func.FuncOp("sum_to", ((register,), (register,))).body) as (
        ub,
    ):
        lb = riscv.LiOp(0)
        step = riscv.LiOp(1)
        initial = riscv.LiOp(0)

        @Builder.implicit_region((register, register, register, register))
        def before_region(args: tuple[BlockArgument, ...]):
            (acc0, i0, ub0, step0) = args
            cond = riscv.SltOp(i0, ub0).rd
            riscv_scf.ConditionOp(cond, acc0, i0, ub0, step0)

        @Builder.implicit_region((register, register, register, register))
        def after_region(args: tuple[BlockArgument, ...]):
            (acc1, i1, ub1, step1) = args
            res = riscv.AddOp(i1, acc1).rd
            i2 = riscv.AddOp(i1, step1).rd
            riscv_scf.YieldOp(res, i2, ub1, step1)

        result, *_ = riscv_scf.WhileOp(
            (initial, lb, ub, step),
            (register, register, register, register),
            before_region,
            after_region,
        ).res
        func.ReturnOp(result)


def interp(module_op: ModuleOp, func_name: str, n: int) -> int:
    module_op.verify()
    interpreter = Interpreter(module_op)
    interpreter.register_implementations(RiscvScfFunctions())
    interpreter.register_implementations(RiscvFunctions())
    interpreter.register_implementations(FuncFunctions())
    (result,) = interpreter.call_op(func_name, (n,))
    return result


@pytest.mark.parametrize("n,res", [(0, 0), (1, 0), (2, 1), (3, 3), (4, 6), (5, 10)])
def test_sum_to(n: int, res: int):
    assert res == interp(sum_to_for_op, "sum_to", n)
    assert res == interp(sum_to_while_op, "sum_to", n)



from __future__ import annotations

import re
from dataclasses import dataclass
from enum import Enum
from string import hexdigits
from typing import Literal, TypeAlias, TypeGuard, cast, overload

from xdsl.utils.exceptions import ParseError
from xdsl.utils.lexer import Lexer, Position, Span, Token

PunctuationSpelling: TypeAlias = Literal[
    "->",
    ":",
    ",",
    "...",
    "=",
    ">",
    "{",
    "(",
    "[",
    "<",
    "-",
    "+",
    "?",
    "}",
    ")",
    "]",
    "*",
    "|",
    "{-#",
    "#-}",
]


@dataclass(frozen=True, repr=False)
class StringLiteral(Span):
    def __post_init__(self):
        if len(self) < 2 or self.text[0] != '"' or self.text[-1] != '"':
            raise ParseError(self, "Invalid string literal!")

    @overload
    @classmethod
    def from_span(cls, span: Span) -> StringLiteral: ...

    @overload
    @classmethod
    def from_span(cls, span: None) -> None: ...

    @classmethod
    def from_span(cls, span: Span | None) -> StringLiteral | None:
        """
        Convert a normal span into a StringLiteral, to facilitate parsing.

        If argument is None, returns None.
        """
        if span is None:
            return None
        return cls(span.start, span.end, span.input)

    @property
    def string_contents(self):
        return self.bytes_contents.decode()

    @property
    def bytes_contents(self) -> bytes:
        bytes_contents = bytearray()
        iter_string = iter(self.text[1:-1])
        for c0 in iter_string:
            if c0 != "\\":
                bytes_contents += c0.encode()
            else:
                c0 = next(iter_string)
                match c0:
                    case "n":
                        bytes_contents += b"\n"
                    case "t":
                        bytes_contents += b"\t"
                    case "\\":
                        bytes_contents += b"\\"
                    case '"':
                        bytes_contents += b'"'
                    case _:
                        c1 = next(iter_string)
                        bytes_contents += int(c0 + c1, 16).to_bytes(1, "big")

        return bytes(bytes_contents)


class MLIRTokenKind(Enum):
    # Markers
    EOF = object()

    # Identifiers
    BARE_IDENT = object()
    """bare-id ::= (letter|[_]) (letter|digit|[_$.])*"""
    AT_IDENT = object()  # @foo
    """at-ident ::= `@` (bare-id | string-literal)"""
    HASH_IDENT = object()  # #foo
    """hash-ident  ::= `#` (digit+ | (letter|[$._-]) (letter|[$._-]|digit)*)"""
    PERCENT_IDENT = object()  # %foo
    """percent-ident  ::= `%` (digit+ | (letter|[$._-]) (letter|[$._-]|digit)*)"""
    CARET_IDENT = object()  # ^foo
    """caret-ident  ::= `^` (digit+ | (letter|[$._-]) (letter|[$._-]|digit)*)"""
    EXCLAMATION_IDENT = object()  # !foo
    """exclamation-ident  ::= `!` (digit+ | (letter|[$._-]) (letter|[$._-]|digit)*)"""

    # Literals
    FLOAT_LIT = object()  # 1.0
    INTEGER_LIT = object()  # 1
    STRING_LIT = object()  # "foo"
    BYTES_LIT = object()  # "foo\00\00"

    # Punctuation
    ARROW = "->"
    COLON = ":"
    COMMA = ","
    ELLIPSIS = "..."
    EQUAL = "="
    GREATER = ">"
    L_BRACE = "{"
    L_PAREN = "("
    L_SQUARE = "["
    LESS = "<"
    MINUS = "-"
    PLUS = "+"
    QUESTION = "?"
    R_BRACE = "}"
    R_PAREN = ")"
    R_SQUARE = "]"
    STAR = "*"
    VERTICAL_BAR = "|"
    FILE_METADATA_BEGIN = "{-#"
    FILE_METADATA_END = "#-}"

    @staticmethod
    def get_punctuation_spelling_to_kind_dict() -> dict[str, MLIRTokenKind]:
        return {
            "->": MLIRTokenKind.ARROW,
            ":": MLIRTokenKind.COLON,
            ",": MLIRTokenKind.COMMA,
            "...": MLIRTokenKind.ELLIPSIS,
            "=": MLIRTokenKind.EQUAL,
            ">": MLIRTokenKind.GREATER,
            "{": MLIRTokenKind.L_BRACE,
            "(": MLIRTokenKind.L_PAREN,
            "[": MLIRTokenKind.L_SQUARE,
            "<": MLIRTokenKind.LESS,
            "-": MLIRTokenKind.MINUS,
            "+": MLIRTokenKind.PLUS,
            "?": MLIRTokenKind.QUESTION,
            "}": MLIRTokenKind.R_BRACE,
            ")": MLIRTokenKind.R_PAREN,
            "]": MLIRTokenKind.R_SQUARE,
            "*": MLIRTokenKind.STAR,
            "|": MLIRTokenKind.VERTICAL_BAR,
            "{-#": MLIRTokenKind.FILE_METADATA_BEGIN,
            "#-}": MLIRTokenKind.FILE_METADATA_END,
        }

    def is_punctuation(self) -> bool:
        punctuation_dict = MLIRTokenKind.get_punctuation_spelling_to_kind_dict()
        return self in punctuation_dict.values()

    @staticmethod
    def is_spelling_of_punctuation(
        spelling: str,
    ) -> TypeGuard[PunctuationSpelling]:
        punctuation_dict = MLIRTokenKind.get_punctuation_spelling_to_kind_dict()
        return spelling in punctuation_dict.keys()

    @staticmethod
    def get_punctuation_kind_from_name(
        spelling: PunctuationSpelling,
    ) -> MLIRTokenKind:
        assert MLIRTokenKind.is_spelling_of_punctuation(spelling), (
            "Kind.get_punctuation_kind_from_name: spelling is not a "
            "valid punctuation spelling!"
        )
        return MLIRTokenKind.get_punctuation_spelling_to_kind_dict()[spelling]

    def get_int_value(self, span: Span):
        """
        Translate the token text into an integer value.
        This function will raise an exception if the token is not an integer
        literal.
        """
        if self != MLIRTokenKind.INTEGER_LIT:
            raise ValueError("Token is not an integer literal!")
        if span.text[:2] in ["0x", "0X"]:
            return int(span.text, 16)
        return int(span.text, 10)

    def get_float_value(self, span: Span):
        """
        Translate the token text into a float value.
        This function will raise an exception if the token is not a float
        literal.
        """
        if self != MLIRTokenKind.FLOAT_LIT:
            raise ValueError("Token is not a float literal!")
        return float(span.text)

    def get_string_literal_value(self, span: Span) -> str:
        """
        Translate the token text into a string literal value.
        This will remove the quotes around the string literal, and unescape
        the string.
        This function will raise an exception if the token is not a string literal.
        """
        if self != MLIRTokenKind.STRING_LIT:
            raise ValueError("Token is not a string literal!")
        return StringLiteral.from_span(span).string_contents


MLIRToken = Token[MLIRTokenKind]


@dataclass
class MLIRLexer(Lexer[MLIRTokenKind]):
    def _is_in_bounds(self, size: Position = 1) -> bool:
        """
        Check if the current position is within the bounds of the input.
        """
        return self.pos + size - 1 < self.input.len

    def _get_chars(self, size: int = 1) -> str | None:
        """
        Get the character at the current location, or multiple characters ahead.
        Return None if the position is out of bounds.
        """
        res = self.input.slice(self.pos, self.pos + size)
        self.pos += size
        return res

    def _peek_chars(self, size: int = 1) -> str | None:
        """
        Peek at the character at the current location, or multiple characters ahead.
        Return None if the position is out of bounds.
        """
        return self.input.slice(self.pos, self.pos + size)

    def _consume_chars(self, size: int = 1) -> None:
        """
        Advance the lexer position in the input by the given amount.
        """
        self.pos += size

    def _consume_regex(self, regex: re.Pattern[str]) -> re.Match[str] | None:
        """
        Advance the lexer position to the end of the next match of the given
        regular expression.
        """
        match = regex.match(self.input.content, self.pos)
        if match is None:
            return None
        self.pos = match.end()
        return match

    _whitespace_regex = re.compile(r"((//[^\n]*(\n)?)|(\s+))*", re.ASCII)

    def _consume_whitespace(self) -> None:
        """
        Consume whitespace and comments.
        """
        self._consume_regex(self._whitespace_regex)

    def _form_token(self, kind: MLIRTokenKind, start_pos: Position) -> MLIRToken:
        """
        Return a token with the given kind, and the start position.
        """
        return MLIRToken(kind, Span(start_pos, self.pos, self.input))

    def lex(self) -> MLIRToken:
        """
        Lex a token from the input, and returns it.
        """
        # First, skip whitespaces
        self._consume_whitespace()

        start_pos = self.pos
        current_char = self._get_chars()

        # Handle end of file
        if current_char is None:
            return self._form_token(MLIRTokenKind.EOF, start_pos)

        # bare identifier
        if current_char.isalpha() or current_char == "_":
            return self._lex_bare_identifier(start_pos)

        # single-char punctuation that are not part of a multi-char token
        single_char_punctuation = {
            ":": MLIRTokenKind.COLON,
            ",": MLIRTokenKind.COMMA,
            "(": MLIRTokenKind.L_PAREN,
            ")": MLIRTokenKind.R_PAREN,
            "}": MLIRTokenKind.R_BRACE,
            "[": MLIRTokenKind.L_SQUARE,
            "]": MLIRTokenKind.R_SQUARE,
            "<": MLIRTokenKind.LESS,
            ">": MLIRTokenKind.GREATER,
            "=": MLIRTokenKind.EQUAL,
            "+": MLIRTokenKind.PLUS,
            "*": MLIRTokenKind.STAR,
            "?": MLIRTokenKind.QUESTION,
            "|": MLIRTokenKind.VERTICAL_BAR,
        }
        if current_char in single_char_punctuation:
            return self._form_token(single_char_punctuation[current_char], start_pos)

        # '...'
        if current_char == ".":
            if self._get_chars(2) != "..":
                raise ParseError(
                    Span(start_pos, start_pos + 1, self.input),
                    "Expected three consecutive '.' for an ellipsis",
                )
            return self._form_token(MLIRTokenKind.ELLIPSIS, start_pos)

        # '-' and '->'
        if current_char == "-":
            if self._peek_chars() == ">":
                self._consume_chars()
                return self._form_token(MLIRTokenKind.ARROW, start_pos)
            return self._form_token(MLIRTokenKind.MINUS, start_pos)

        # '{' and '{-#'
        if current_char == "{":
            if self._peek_chars(2) == "-#":
                self._consume_chars(2)
                return self._form_token(MLIRTokenKind.FILE_METADATA_BEGIN, start_pos)
            return self._form_token(MLIRTokenKind.L_BRACE, start_pos)

        # '#-}'
        if current_char == "#" and self._peek_chars(2) == "-}":
            self._consume_chars(2)
            return self._form_token(MLIRTokenKind.FILE_METADATA_END, start_pos)

        # '@' and at-identifier
        if current_char == "@":
            return self._lex_at_ident(start_pos)

        # '#', '!', '^', '%' identifiers
        if current_char in ["#", "!", "^", "%"]:
            return self._lex_prefixed_ident(start_pos)

        if current_char == '"':
            return self._lex_string_literal(start_pos)

        if current_char.isnumeric():
            return self._lex_number(start_pos)

        raise ParseError(
            Span(start_pos, start_pos + 1, self.input),
            f"Unexpected character: {current_char}",
        )

    IDENTIFIER_SUFFIX = r"[a-zA-Z0-9_$.]*"
    bare_identifier_regex = re.compile(r"[a-zA-Z_]" + IDENTIFIER_SUFFIX)
    bare_identifier_suffix_regex = re.compile(IDENTIFIER_SUFFIX)

    def _lex_bare_identifier(self, start_pos: Position) -> MLIRToken:
        """
        Lex a bare identifier with the following grammar:
        `bare-id ::= (letter|[_]) (letter|digit|[_$.])*`

        The first character is expected to have already been parsed.
        """
        self._consume_regex(self.bare_identifier_suffix_regex)

        return self._form_token(MLIRTokenKind.BARE_IDENT, start_pos)

    def _lex_at_ident(self, start_pos: Position) -> MLIRToken:
        """
        Lex an at-identifier with the following grammar:
        `at-id ::= `@` (bare-id | string-literal)`

        The first character `@` is expected to have already been parsed.
        """
        current_char = self._get_chars()

        if current_char is None:
            raise ParseError(
                Span(start_pos, start_pos + 1, self.input),
                "Unexpected end of file after @.",
            )

        # bare identifier case
        if current_char.isalpha() or current_char == "_":
            token = self._lex_bare_identifier(start_pos)
            return self._form_token(MLIRTokenKind.AT_IDENT, token.span.start)

        # literal string case
        if current_char == '"':
            token = self._lex_string_literal(start_pos)
            return self._form_token(MLIRTokenKind.AT_IDENT, token.span.start)

        raise ParseError(
            Span(start_pos, self.pos, self.input),
            "@ identifier expected to start with letter, '_', or '\"'.",
        )

    _suffix_id = re.compile(r"([0-9]+|[a-zA-Z$._-][a-zA-Z0-9$._-]*)")

    def _lex_prefixed_ident(self, start_pos: Position) -> MLIRToken:
        """
        Parsed the following prefixed identifiers:
        ```
        hash-ident  ::= `#` suffix-id
        percent-ident  ::= `%` suffix-id
        caret-ident  ::= `^` suffix-id
        exclamation-ident  ::= `!` suffix-id
        ```
        with
        ```
        suffix-id = (digit+ | (letter|[$._-]) (letter|[$._-]|digit)*)
        ```

        The first character is expected to have already been parsed.
        """
        assert self.pos != 0, (
            "First prefixed identifier character must have been parsed"
        )
        first_char = self.input.at(self.pos - 1)
        if first_char == "#":
            kind = MLIRTokenKind.HASH_IDENT
        elif first_char == "!":
            kind = MLIRTokenKind.EXCLAMATION_IDENT
        elif first_char == "^":
            kind = MLIRTokenKind.CARET_IDENT
        else:
            assert first_char == "%", (
                "First prefixed identifier character must have been parsed correctly"
            )
            kind = MLIRTokenKind.PERCENT_IDENT

        match = self._consume_regex(self._suffix_id)
        if match is None:
            raise ParseError(
                Span(start_pos, self.pos, self.input),
                "Expected suffix identifier after {first_char}",
            )

        return self._form_token(kind, start_pos)

    _unescaped_characters_regex = re.compile(r'[^"\\\n\v\f]*')

    def _lex_string_literal(self, start_pos: Position) -> MLIRToken:
        """
        Lex a string literal.
        The first character `"` is expected to have already been parsed.
        """

        bytes_token = False
        while self._is_in_bounds():
            self._consume_regex(self._unescaped_characters_regex)
            current_char = self._get_chars()

            # end of string literal
            if current_char == '"':
                if bytes_token:
                    return self._form_token(MLIRTokenKind.BYTES_LIT, start_pos)
                else:
                    return self._form_token(MLIRTokenKind.STRING_LIT, start_pos)

            # newline character in string literal (not allowed)
            if current_char in ["\n", "\v", "\f"]:
                raise ParseError(
                    Span(start_pos, self.pos, self.input),
                    "Newline character not allowed in string literal.",
                )

            # escape character
            # TODO: handle unicode escape
            if current_char == "\\":
                escaped_char = self._get_chars()
                if escaped_char not in ['"', "\\", "n", "t"]:
                    bytes_token = True
                    next_char = self._get_chars()
                    if escaped_char is None or next_char is None:
                        raise ParseError(
                            Span(start_pos, self.pos, self.input),
                            "Unknown escape in string literal.",
                        )
                    try:
                        int(escaped_char + next_char, 16)
                    except Exception:
                        raise ParseError(
                            Span(start_pos, self.pos, self.input),
                            "Unknown escape in string literal.",
                        )

        raise ParseError(
            Span(start_pos, self.pos, self.input),
            "End of file reached before closing string literal.",
        )

    _hexdigits_star_regex = re.compile(r"[0-9a-fA-F]*")
    _digits_star_regex = re.compile(r"[0-9]*")
    _fractional_suffix_regex = re.compile(r"\.[0-9]*([eE][+-]?[0-9]+)?")

    def _lex_number(self, start_pos: Position) -> MLIRToken:
        """
        Lex a number literal, which is either a decimal or an hexadecimal.
        The first character is expected to have already been parsed.
        """
        first_digit = self.input.at(self.pos - 1)

        # Hexadecimal case, we only parse it if we see the first '0x' characters,
        # and then a first digit.
        # Otherwise, a string like '0xi32' would not be parsed correctly.
        if (
            first_digit == "0"
            and self._peek_chars() == "x"
            and self._is_in_bounds(2)
            and cast(str, self.input.at(self.pos + 1)) in hexdigits
        ):
            self._consume_chars(2)
            self._consume_regex(self._hexdigits_star_regex)
            return self._form_token(MLIRTokenKind.INTEGER_LIT, start_pos)

        # Decimal case
        self._consume_regex(self._digits_star_regex)

        # Check if we are lexing a floating point
        match = self._consume_regex(self._fractional_suffix_regex)
        if match is not None:
            return self._form_token(MLIRTokenKind.FLOAT_LIT, start_pos)
        return self._form_token(MLIRTokenKind.INTEGER_LIT, start_pos)



from xdsl.dialects.builtin import (
    I32,
    I64,
    AnyFloat,
    BoolAttr,
    DenseArrayBase,
    FloatAttr,
    IntegerAttr,
    TensorType,
)
from xdsl.ir import Dialect
from xdsl.irdl import (
    IRDLOperation,
    ParsePropInAttrDict,
    irdl_op_definition,
    operand_def,
    prop_def,
    result_def,
)


@irdl_op_definition
class ClampOp(IRDLOperation):
    """
    Computes clamp(features, min, max)
    """

    name = "tosa.clamp"

    min_int = prop_def(IntegerAttr[I64])
    max_int = prop_def(IntegerAttr[I64])

    min_fp = prop_def(FloatAttr[AnyFloat])
    max_fp = prop_def(FloatAttr[AnyFloat])

    input = operand_def(TensorType)
    output = result_def(TensorType)

    irdl_options = [ParsePropInAttrDict()]

    assembly_format = "$input attr-dict `:` `(` type($input) `)` `->` type($output)"


@irdl_op_definition
class RescaleOp(IRDLOperation):
    """
    Tosa Rescale Operator
    """

    name = "tosa.rescale"

    input_zp = prop_def(IntegerAttr[I32])
    output_zp = prop_def(IntegerAttr[I32])
    multiplier = prop_def(DenseArrayBase)
    shift = prop_def(DenseArrayBase)
    scale32 = prop_def(BoolAttr)
    double_round = prop_def(BoolAttr)
    per_channel = prop_def(BoolAttr)

    input = operand_def(TensorType)
    output = result_def(TensorType)

    irdl_options = [ParsePropInAttrDict()]

    assembly_format = "$input attr-dict `:` `(` type($input) `)` `->` type($output)"


TOSA = Dialect(
    "tosa",
    [
        ClampOp,
        RescaleOp,
    ],
    [],
)



from dataclasses import dataclass, field
from typing import cast

from xdsl.context import Context
from xdsl.dialects import riscv, riscv_func
from xdsl.dialects.builtin import ModuleOp
from xdsl.ir import Operation
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)


class LowerSyscallOp(RewritePattern):
    """
    Lower SSA version of syscall, storing the optional result to a0.


    Different platforms have different calling conventions. This lowering assumes that
    the inputs are stored in a0-a6, and the opcode is stored to a7. Upon return, the
    a0 contains the result value. This is not the case for some kernels.

    In the future, this pass should take the compilation target as a parameter to guide
    the rewrites.

    Issue tracking this: https://github.com/xdslproject/xdsl/issues/952
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: riscv_func.SyscallOp, rewriter: PatternRewriter):
        ops: list[Operation] = []

        for i, arg in enumerate(op.args):
            ops.append(
                riscv.MVOp(
                    arg,
                    rd=riscv.IntRegisterType.a_register(i),
                )
            )

        ops.append(riscv.LiOp(immediate=op.syscall_num, rd=riscv.Registers.A7))

        if op.result is None:
            ops.append(riscv.EcallOp())
            new_results = ()
        else:
            # The result will be stored to a0, move to register that will be used
            ecall = riscv.EcallOp()
            ops.append(ecall)
            gr = riscv.GetRegisterOp(riscv.Registers.A0)
            ops.append(gr)
            res = gr.res

            mv = riscv.MVOp(res, rd=cast(riscv.IntRegisterType, op.result.type))
            ops.append(mv)
            new_results = mv.results

        rewriter.replace_matched_op(ops, new_results=new_results)


class InsertExitSyscallOp(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: riscv_func.ReturnOp, rewriter: PatternRewriter):
        parent_op = op.parent_op()
        if (
            not isinstance(parent_op, riscv_func.FuncOp)
            or parent_op.sym_name.data != "main"
        ):
            return

        EXIT = 93
        rewriter.insert_op_before_matched_op(riscv_func.SyscallOp(EXIT))


@dataclass(frozen=True)
class LowerRISCVFunc(ModulePass):
    name = "lower-riscv-func"

    insert_exit_syscall: bool = field(default=False)

    def apply(self, ctx: Context, op: ModuleOp) -> None:
        if self.insert_exit_syscall:
            PatternRewriteWalker(
                InsertExitSyscallOp(), apply_recursively=False
            ).rewrite_module(op)
        PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    LowerSyscallOp(),
                ]
            )
        ).rewrite_module(op)



from __future__ import annotations

from collections.abc import Callable, Iterable
from contextlib import contextmanager
from dataclasses import dataclass, field
from typing import IO, Any, TypeVar


@dataclass(eq=False, repr=False)
class BasePrinter:
    stream: IO[str] | None = field(default=None)
    indent_num_spaces: int = field(default=2, kw_only=True)
    _indent: int = field(default=0, init=False)
    _current_line: int = field(default=0, init=False)
    _current_column: int = field(default=0, init=False)

    _next_line_callback: list[Callable[[], None]] = field(
        default_factory=list, init=False
    )

    def print_string(self, text: str, *, indent: int | None = None) -> None:
        """
        Prints a string to the printer's output.

        This function takes into account indentation level when
        printing new lines.
        If the indentation level is specified as 0, the string is printed as-is, if `None`
        then the `Printer` instance's indentation level is used.
        """

        num_newlines = text.count("\n")

        if not num_newlines:
            self._current_column += len(text)
            print(text, end="", file=self.stream)
            return

        indent = self._indent if indent is None else indent
        lines = text.split("\n")

        if indent == 0 and not self._next_line_callback:
            # No indent and no callback to print after the next newline, the text
            # can be printed directly.
            self._current_line += num_newlines
            self._current_column = len(lines[-1])
            print(text, end="", file=self.stream)
            return

        # Line and column information is not computed ahead of time
        # as indent-aware newline printing may use it as part of
        # callbacks.
        print(lines[0], end="", file=self.stream)
        self._current_column += len(lines[0])
        for line in lines[1:]:
            self._print_new_line(indent=indent)
            print(line, end="", file=self.stream)
            self._current_column += len(line)

    T = TypeVar("T")

    def print_list(
        self, elems: Iterable[T], print_fn: Callable[[T], Any], delimiter: str = ", "
    ) -> None:
        for i, elem in enumerate(elems):
            if i:
                self.print_string(delimiter)
            print_fn(elem)

    def _print_new_line(
        self, indent: int | None = None, print_message: bool = True
    ) -> None:
        indent = self._indent if indent is None else indent
        # Prints a newline, bypassing the `print_string` method
        print(file=self.stream)
        self._current_line += 1
        if print_message:
            for callback in self._next_line_callback:
                callback()
            self._next_line_callback = []
        num_spaces = indent * self.indent_num_spaces
        # Prints indentation, bypassing the `print_string` method
        print(" " * num_spaces, end="", file=self.stream)
        self._current_column = num_spaces

    @contextmanager
    def indented(self, amount: int = 1):
        """
        Increases the indentation level by the provided amount
        for the duration of the context.

        Only affects new lines printed within the context.
        """

        self._indent += amount
        try:
            yield
        finally:
            self._indent -= amount

    def _add_message_on_next_line(self, message: str, begin_pos: int, end_pos: int):
        """Add a message that will be displayed on the next line."""

        def callback(indent: int = self._indent):
            self._print_message(message, begin_pos, end_pos, indent)

        self._next_line_callback.append(callback)

    def _print_message(
        self, message: str, begin_pos: int, end_pos: int, indent: int | None = None
    ):
        """
        Print a message.
        This is expected to be called at the beginning of a new line and to create a new
        line at the end.
        The span of the message to be underlined is represented as [begin_pos, end_pos).
        """
        indent = self._indent if indent is None else indent
        indent_size = indent * self.indent_num_spaces
        self.print_string(" " * indent_size)
        message_end_pos = max(map(len, message.split("\n"))) + indent_size + 2
        first_line = (
            (begin_pos - indent_size) * "-"
            + (end_pos - begin_pos) * "^"
            + (max(message_end_pos, end_pos) - end_pos) * "-"
        )
        self.print_string(first_line)
        self._print_new_line(indent=indent, print_message=False)
        for message_line in message.split("\n"):
            self.print_string("| ")
            self.print_string(message_line)
            self._print_new_line(indent=indent, print_message=False)
        self.print_string("-" * (max(message_end_pos, end_pos) - indent_size))
        self._print_new_line(indent=0, print_message=False)



import marimo

__generated_with = "0.11.16"
app = marimo.App(width="medium")


@app.cell(hide_code=True)
def _():
    import marimo as mo

    from xdsl.irdl import (
        irdl_op_definition,
        irdl_attr_definition,
        ParametrizedAttribute,
        IRDLOperation,
        attr_def,
        result_def,
        operand_def,
        region_def,
        traits_def,
    )
    return (
        IRDLOperation,
        ParametrizedAttribute,
        attr_def,
        irdl_attr_definition,
        irdl_op_definition,
        mo,
        operand_def,
        region_def,
        result_def,
        traits_def,
    )


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        # The Multi-Level Approach to Compilation

        Most programming languages are rather far removed from assembly code. As a result, translating a language directly into Assembly is both difficult (easy to make mistakes) and inefficient (generated Assembly code is slow!). Therefore, most compilers go through many intermediate representations / languages (IRs), which represent different layers of abstractions, to provide an efficient translation down to Assembly code - which is also feasible for compiler engineers to write.

        Although an effective approach, over the years it has become apparent that many compilers are repeatedly defining similar intermediate languages. Given the separate architecture of different compilers, it has been quite difficult to reuse code between architectures.

        Previous projects such as LLVM have provided a common abstraction layer for different compilers to reuse - however this is only one layer, and compilers of different domains are still having to repeat significant amounts of work to reach the LLVM level.

        LLVM is also not necessarily the ideal layer to perform optimisations on - thus many compilers will also implement their own implementations on custom IRs before they reach LLVM.

        xDSL is a framework that makes it easy to:

        1. Define custom abstraction layers for compilers
        2. Reuse abstractions from other compilers
        3. Encode domain-specific knowledge (i.e. optimisations) within these abstraction layers

        xDSL takes inspiration from MLIR, which is a framework based upon similar ideas, and is implemented in C++.
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        # General Concepts

        The kind of IRs that xDSL builds are SSA-based - a restriction placed on IRs which makes it easy for analyzing the code.
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        ### Static Single Assignment (SSA)
        The kind of programming languages xDSL defines have the $\textit{static single assignment}$ (SSA) property. It is a property on variables that means they can be assigned a value only once, and once assigned a value, they cannot be modified. This is a common restriction used within compilers, as it enables many compiler optimisations to be performed.

        In an SSA-based language, values are generated by the language's constructs, and old values may be referred to in the construction of new values.

        <!-- give an example of SSA code snippet -->

        Further reading: https://en.wikipedia.org/wiki/Static_single-assignment_form
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""# xDSL""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        ## What is an IR, anyway?

        xDSL is a framework for modelling/specifying IRs. Before we go into the details of xDSL, it is perhaps useful to take a step back and consider what exactly an IR is!

        Code can be represented in a human-readable textual format - however strings are difficult for compilers to handle. An IR is an *equivalent*, structured representation of code convenient for machines to parse through. Importantly, it is solely a representation of the code - is not directly executable on a processor.

        Languages are often designed for particular tasks and domains, which allows certain optimisations or types of reasoning to take place more easily within the language. A compiler might want to optimise the code in many different ways, and therefore can utilise multiple languages (represented as IR within the compiler) during its translation into machine code.

        Wikipedia has a nice article as a starting point for further reading: https://en.wikipedia.org/wiki/Intermediate_representation
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        ## Programming Language Syntax
        All programming languages will have a set of "language constructs", or, the _vocabulary_ available within the programming language.

        These constructs can be composed together, forming "expressions" - which are sentences that _could_ be evaluated to a value.

        Just like in English, though, vocabulary (words) cannot be composed together arbitrarily - they should follow the syntactical rules of the language. For programming languages, this is known as the _syntax_ of the language, and is often specified in terms of a [grammar](https://en.wikipedia.org/wiki/Formal_grammar).

        Here are some examples of building blocks within various languages:

        - Arithmetic: a language that describes arithmetic might have constructs, including:
            - `+, -, x, /, round, max`
            - The values of this language are real numbers (floats, say).
            - One syntax rule in this language is that each construct should take in either values, or other valid expressions.
        - Python: includes too many constructs to fit in this code block, but some examples include:
            - `for, if, while, assert, def, class, import`
            - Since Python is object-oriented, values in Python language are objects.
            - A valid expression in Python would follow the [Python grammar](https://docs.python.org/3/reference/grammar.html), a rather large document specifying all syntax rules of Python.
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        ## Programming Language Semantics
        In English, a sentence being grammatically correct is not enough for it to have _meaning_. The meaning of sentences is embedded in the intuition of English speakers, and speakers need to generally agree on the meaning (semantics) of sentences in order to communicate.

        Similarly, for programming languages, the grammar is not enough to convey meaning. There needs to be some semantics of the language which is generally agreed across the users of the programming language and the compiler implementations.

        Programmers make use of the language semantics to write programs which do what they think it does, and compilers utilize the semantics to optimize the programs whilst preserving their semantics - after all, the transformation should not change the meaning of what the programmer wrote!

        For programming languages, the semantics could be assigned to a mathematical function, unambiguously defining the meaning for all sentences in the programming languages.

        However, in industrial programming languages, semantics is often an intuitive notion that is embedded in the minds of language users and within the compiler's transformations (which is preserved).
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        ## Programming Language

        With the above two definitions, a programming language's definition simply consists of two parts:

        1. the syntax
        2. the semantics
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        ## xDSL: A Tool for Designing and Implementing Languages for Compilers

        Following from the previous definition of a programming language, the corresponding parts of a compiler IR consists of:

        - The representation of the IR (i.e. the syntax - what constructs there are, how do they compose, what is the data structure to use in the implementation?)
        - The transformations performed on the IR (i.e. the implied semantics: utilizing the expressive powers of the IR in its domain for optimizations)

        xDSL provides the tools for modelling IRs, where:

        - Modelling the IR's syntax is done via defining the _operations_, and _attributes_ within the dialect
        - Modelling the IR's transformation is done via specifying _rewrite patterns_ on the IR

        To be able to specify the syntax and the transformations of any IR, it needs to be general enough to fit the wide range of potential designs and uses of language constructs.

        Below, we will introduce the concepts used within xDSL.
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        ### Operations: Modelling General Language Constructs

        Operations are units used within xDSL to model a "language construct", i.e. a vocabulary available within the language.

        A set of operations together would then describe the syntax of the language.

        An operation is the combination of:   <!-- make it possible to click on the links to go to respective blocks perhaps -->

        1. A name: the name of the operation
        2. A (possibly empty) list of _regions_
        3. A (possibly empty) list of _operands_
        4. A (possibly empty) set of _attributes_
        5. A (nonempty) list of results

        Operations would take in some previously defined SSA values in the program (as operands), and return as a result new SSA value(s).

        The above definition of an operation may seem rather terse - this is because an operation is designed to be general enough to describe all kinds of constructs within a programming language.

        And so we will provide examples of operations below for more intuition. Moreover, please refer to the individual definitions for more information.
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        ### Regions

        A _region_ represents a sequence of instructions with potentially non-linear control flow.

        That is, within a region, the program could execute not just from top-to-bottom, but it may go into different branches or loops. A region therefore corresponds to a control flow graph (_CFG_) - where the nodes are blocks, and edges are the flow of control: we have an edge when a block passes control to another block.

        In xDSL, a _region_ consists of:

        1. A list of _blocks_
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        ### Blocks

        As commonly seen in compiler theory, a block represents a sequence of instructions that must execute from top-to-bottom, one after the other.

        At the end of the block, the control flow is handed over to one of its potential _successors_ (which one exactly is dependent on the precise values computed, which is done at runtime)

        What's different about blocks in xDSL is that blocks contain operations, which themselves could contain regions, and regions can contain blocks once again.

        Thus, the structure of the IR is recursive within xDSL.

        In xDSL, a block is the combination of:

        1. A list of _block arguments_
        2. A list of _operations_
        3. A (possibly empty) list of _successor_ _blocks_
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        ### Block Arguments

        Instead of using [$\phi$ nodes](https://www.cs.princeton.edu/~appel/papers/ssafun.pdf), xDSL uses _block arguments_ to decide the value to use for computation when input values could come from several different branches.

        These are two equivalent ways to solve the same problem, though, as explained in the earlier linked [article](https://www.cs.princeton.edu/~appel/papers/ssafun.pdf) on $\phi$ nodes.

        In xDSL, block arguments is the combination of:

        1. An index, denoting the argument's position within the block
        2. A type, i.e. the type of the argument
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        ### Operands

        Operands are the arguments to an operation. This allows the operation to use previously calculated values, which is essential for imperative programs.

        Note that the operands are different to block arguments in that:

        - _operands_ within an operation refers to a unique previously defined SSA values, whereas
        - _block arguments_ are more akin to function arguments, and do not specify a unique previously defined SSA value to use

        In xDSL, operands are defined as:

        1. An SSA value: that is, the previously defined SSA value used in this operand
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        ### Attributes

        Attributes are data, in a predetermined format, which are associated with operations. The format of attributes for a particular language come from the language specification.

        The data it stores must be information which is known at compile time.

        There are several uses of attributes with xDSL, some examples include:

        - **_type information_**: in xDSL, the types of operation's results are stored as attributes
        - storage for code analysis: [data flow analysis](https://clang.llvm.org/docs/DataFlowAnalysisIntro.html) passes computes properties such as liveness and range of variables, which relies on being able to store and update information within the IR
        - operations whose semantics is dependent on attributes: such as
            - [convolution operations](https://mlir.llvm.org/docs/Dialects/TOSA/#tosaconv2d-mlirtosaconv2dop) (semantics depends on its stride and padding attributes),
            - [comparison operations](https://mlir.llvm.org/docs/Dialects/ArithOps/#arithcmpi-mlirarithcmpiop) (semantics depends on its "predicate" attribute, which specifies exactly what comparison to make, i.e. signed less than, unsigned greater than or equals, and so on)
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        ### Dialects

        Languages in xDSL are known as `dialects` - which is a grouping of related operations and attributes. For instance,

        - The `scf` (structured control flow) dialect groups together
            - Operations such as: for/while loops, reduce, yield, if/else statements
            - Attributes such as: no new attributes in this dialect! It instead uses attributes/types from _other_ dialects
        - The `arith` (arithmetic) dialect groups together
            - Operation such as: comparison, integer/floating-point addition/multiplication, or/xor/and logical bitwise operations, and so on
            - Attributes such as: predicates to decide the mode of comparison, flags to determine mode of floating point computation used
        - The `async` (asynchronous) dialect groups together operations commonly seen in async applications, as well as associated attributes (types).
            - Operations such as: asynchronous calls, waits, async functions
            - Attributes such as: coroutine reference types, future types

        Dialects can be _combined_, such that the combined dialect have the operations and attributes from all constituent dialects.
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    _snippet = '''\
    class RewritePattern(ABC):
        \"""
        A side-effect free rewrite pattern matching on a DAG.
        \"""

        @abstractmethod
        def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter):
            \"""
            Match an operation, and optionally perform a rewrite using the rewriter.
            \"""
            # here, determine if the operation matches our pattern
            # and define a transformation to be applied on the IR making use of the
            # `rewrite` as an API for changing the IR.
            ...
    '''

    mo.md(
        fr"""
        ### Rewrite Patterns

        Rewrite pattern is a tool to transform IRs defined using xDSL.

        It makes it easy to specify peephole optimizations. Rewrite patterns specify a _source_ pattern, and a transformation, which applies itself to the IR by:

        - matching a _source_ type pattern within the IR
        - updating the IR by applying the predefined transformation

        Multiple rewrite patterns can be applied and composed together in compiler passes.

        A function signature is worth a thousand words... :)

        {mo.ui.code_editor(_snippet, disabled=True)}
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""# xDSL Examples""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        ## 1. An SQL Dialect

        As a first example, we'll describe an SSA IR for describing SQL queries, with ability to query and filter on tables. This will be a simple IR which covers two things:
        * reading from tables (queries like `%table = select(table_name)`)
        * filtering results (`%filtered_table = filter(%table, {filter_conditions})`)

        Each `select` and `filter` constructs returns an object of `Bag` type, representing a query's results.

        The results of the query/filtering operations are stored in SSA variables (whose names are prefixed with a `%`)

        The language allows us to perform SQL queries - for instance, the one below:
        ```SQL
        SELECT * FROM T WHERE T.a > 5 + 5
        ```

        would look something like this in our SSA IR for SQL:
        ```
        %table = select("T")
        %filtered = filter(%table, (lambda t: t.a > 5 + 5))
        ```

        A compiler for our language should be able to do some optimizations on the code as well. For instance, the above SSA form could be transformed like so:
        ```
        %table = select("T")
        %filtered = filter(%table, (lambda t: t.a > 10))
        ```
        where the constants `5+5` have been folded.
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        ### xDSL Modelling

        Now let's consider how we would model the IR with xDSL.

        The two features of the IR, `select` and `filter`, can be modelled as _operations_: they take in _operands_, and return an SSA value - fitting operations very well.

        Let's start the modelling in xDSL with what we've decided so far:
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    example_code_text1 = """@irdl_op_definition
    class SelectOp(IRDLOperation):
        name = "sql.select"
        ...

    @irdl_op_definition
    class FilterOp(IRDLOperation):
        name = "sql.filter"
        ...
    """

    mo.md(
        rf"""
    {mo.ui.code_editor(example_code_text1, language="python", disabled=True)}
    """
    )
    return (example_code_text1,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        In xDSL, the dialect definitions are made easy through a framework called IRDL - which reduces much of the boilerplate when defining operations.

        The operation tooling is implemented as Python decorators, as seen in the `@irdl_op_definition` annotation.

        As mentioned above, the dialect introduces a new type, called `Bag`, which represents the return value from queries.

        In xDSL, types can be implemented as _attributes_ like so:
        """
    )
    return


@app.cell
def _(ParametrizedAttribute, irdl_attr_definition):
    @irdl_attr_definition
    class Bag(ParametrizedAttribute):
        name = "sql.bag"
    return (Bag,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        Once again, here we make use of IRDL to help us define attributes, as seen in the `@irdl_attr_definition` annotation.

        With our types, defined, we can now specify the _operands_ for our operations:
        """
    )
    return


@app.cell
def _(Bag, IRDLOperation, attr_def, irdl_op_definition, result_def):
    from xdsl.dialects.builtin import StringAttr  # (1)


    @irdl_op_definition
    class SelectOp(IRDLOperation):
        name = "sql.select"
        table_name = attr_def(StringAttr)  # (1)
        result_bag = result_def(Bag)  # (3)
    return SelectOp, StringAttr


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        Here comments `#` denotes the new lines we added to the _operation_ definitions.

        Let's go through each of these additions:

        1. To specify the names of our table (which is a string), we needed a string type. Taking advantage of the xDSL ecosystem, we can reuse other dialects which already have similar features defined.
            - Indeed, the `builtin` dialect already has a `StringAttr` attribute defined, and so we use it here for the name of our table.
            - Notice we have made the choice to model the table name in the `select(table_name)` operation as an _attribute_ (as seen by `OpAttr[StringAttr]` annotation). This means that table names in our IR must be known at compile time. This is just a design choice we've made, and it's possible to have the `table_name` be an _operand_ instead, which is an SSA value, and thus doesn't need to be known at compile time.
        2. We specify the _operands_ of the `filter` operation. Here we are saying that `sql.filter` takes in an SSA value.
            - We are additionally saying that the SSA value needs to be of type `Bag`, which enables type-check capabilities automatically for our IRs, a feature provided by xDSL.
        3. We specify the result of our operations - both of our operations should return another SSA value.
            - We are also putting type constraints on our return types, this allows xDSL to type-check our IRs.
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        We're not quite done yet though. Or `sql.filter` operation doesn't yet model the function used for the filter:

        ```
        filter(%table, (lambda t: t.a > 5 + 5))
        ```

        The filter can be modelled as a _region_, which, recalling from earlier, is a list of `blocks`. Since a region corresponds to a control-flow graph, it is suitable for representing functions such as the one above.

        So, our `sql.filter` operation now looks like this:
        """
    )
    return


@app.cell
def _(
    Bag,
    IRDLOperation,
    irdl_op_definition,
    operand_def,
    region_def,
    result_def,
    traits_def,
):
    from xdsl.traits import NoTerminator


    @irdl_op_definition
    class FilterOp(IRDLOperation):
        name = "sql.filter"
        input_bag = operand_def(Bag)
        filter = region_def()
        result_bag = result_def(Bag)
        traits = traits_def(NoTerminator())
    return FilterOp, NoTerminator


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""### Sample Code in our SQL Language""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        Now let's see how the original SSA code:

        ```
        %table = select("T")
        %filtered = filter(%table, (lambda t: t.a > 5 + 5))
        ```

        would be represented in our IR:
        """
    )
    return


@app.cell
def _(Bag, FilterOp, SelectOp):
    from xdsl.printer import Printer  # xDSL tool to visualize our IR
    from xdsl.builder import ImplicitBuilder
    from xdsl.ir import Block, Region

    # getting parts of other dialects defined elsewhere
    from xdsl.dialects import builtin, arith

    printer = Printer()

    # initializes a "SelectOp" operation:
    # select("T")
    table = SelectOp.build(
        attributes={"table_name": builtin.StringAttr("T")}, result_types=[Bag()]
    )
    table.verify()  # performs type checking and structural checking (e.g. based on operation definition). This will throw if types are invalid.
    printer.print_op(table)

    # defines a filter region                                     # (1)

    block = Block(arg_types=(builtin.i32,))  # filter argument
    with ImplicitBuilder(block) as (arg0,):
        # arg0 is filter argument
        # all operations created in an ImplicitBuilder context will be added to its block
        const1 = arith.ConstantOp.from_int_and_width(5, 32)
        const2 = arith.ConstantOp.from_int_and_width(5, 32)
        summed = arith.AddiOp(const1, const2)
        # sgt stands for `signed greater than`. xDSL represents internally as a constant attribute "4"
        cmp_result = arith.CmpiOp(arg0, summed, "sgt")
    filter_region = Region(block)

    # initializes a "FilterOp" operation using previous selection result and defined region:
    # filter(%table, (lambda t: t > 5 + 5))
    filtered = FilterOp.build(
        result_types=[Bag()], operands=[table], regions=[filter_region]
    )
    filtered.verify()
    printer.print_op(filtered)
    return (
        Block,
        ImplicitBuilder,
        Printer,
        Region,
        arg0,
        arith,
        block,
        builtin,
        cmp_result,
        const1,
        const2,
        filter_region,
        filtered,
        printer,
        summed,
        table,
    )


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        We get something like this printed:
        ```
        %0 : !sql.bag = sql.select() ["table_name" = "T"]
        %1 : !sql.bag = sql.filter(%0 : !sql.bag) {
        ^0(%2 : !i32):
          %3 : !i32 = arith.constant() ["value" = 5 : !i32]
          %4 : !i32 = arith.constant() ["value" = 5 : !i32]
          %5 : !i32 = arith.addi(%3 : !i32, %4 : !i32)
          %6 : !i1 = arith.cmpi(%2 : !i32, %5 : !i32) ["predicate" = 4 : !i64]
        }
        ```

        Note that we've mixed in several other dialects here:
        - the `arith` dialect for its numerical operations (`arith.constant`, `arith.addi`, `arith.cmpi`)
        - the `builtin` dialect for its numerical types (`builtin.i32`)

        xDSL generates a `.verify()` function which validates the structural (right number of operands, results, regions), and typing (specified in operation definition) constraints of the operation. Custom verification functions can also be implemented by overriding the `.verify_()` function within the operation definition.

        The filter region represents the lambda `\a -> a > 5+5` by utilizing a single block:
        - We constructed the block by the builder function `Block.from_callable`
        - The argument of the `lambda` is the block arguments
        - The computation of the `lambda` is represented by the block's contents: a list of operations
        - The final operation within a block is a special one. It must be an operation to either:
            - pass control flow to one of the block's _successors_, or
            - be a terminator operation like `scf.yield`, which is akin to a "return".

        So we now have a representation of the original code as an xDSL IR (which are objects in memory).

        This representation allows us to transform and optimize our IR. We'll see a few examples in the section below.
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""### Transforming the xDSL IR""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        By inspecting the IR, we notice that the results of the addition/comparison operations within the `lambda` are already known at compile-time - since their inputs are all known.

        So, we can write a transformation to "fold" the constants to cut down the runtime work.

        We will implement the constant folding transformation as an xDSL `RewritePattern`, and apply this to our IR using xDSL's _RewriteEngine_.
        """
    )
    return


@app.cell
def _(arith):
    from xdsl.pattern_rewriter import (
        PatternRewriter,
        RewritePattern,
        op_type_rewrite_pattern,
    )


    class ConstantFolding(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, op: arith.AddiOp, rewriter: PatternRewriter):
            # pattern: if both arguments to the Addi operation are from `Constant` operations
            if isinstance(op.lhs.op, arith.ConstantOp) and isinstance(
                op.rhs.op, arith.ConstantOp
            ):
                # transform: replace the operation by calculating the sum of the constants at compile time
                return rewriter.replace_matched_op(
                    arith.ConstantOp.from_int_and_width(
                        op.lhs.op.value.value.data + op.rhs.op.value.value.data,
                        op.lhs.op.value.type.width.data,
                    )
                )
    return (
        ConstantFolding,
        PatternRewriter,
        RewritePattern,
        op_type_rewrite_pattern,
    )


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""applying this to our IR:""")
    return


@app.cell
def _(ConstantFolding, filtered, printer):
    from xdsl.pattern_rewriter import (
        PatternRewriteWalker,
        GreedyRewritePatternApplier,
    )


    walker1 = PatternRewriteWalker(
        GreedyRewritePatternApplier([ConstantFolding()]),
        walk_regions_first=True,
        apply_recursively=True,
        walk_reverse=False,
    )
    walker1.rewrite_region(filtered.filter)
    printer.print_op(filtered)
    return GreedyRewritePatternApplier, PatternRewriteWalker, walker1


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""the above introduced some variables which are no longer used (e.g. `%3` and `%4`), therefore can be safely removed from the code.""")
    return


@app.cell
def _(
    GreedyRewritePatternApplier,
    PatternRewriteWalker,
    PatternRewriter,
    RewritePattern,
    arith,
    filtered,
    op_type_rewrite_pattern,
    printer,
):
    class DeadConstantElim(RewritePattern):
        @op_type_rewrite_pattern
        def match_and_rewrite(self, op: arith.ConstantOp, rewriter: PatternRewriter):
            if len(op.result.uses) == 0:
                rewriter.erase_matched_op()


    walker2 = PatternRewriteWalker(
        GreedyRewritePatternApplier([DeadConstantElim()]),
        walk_regions_first=True,
        apply_recursively=True,
        walk_reverse=False,
    )
    walker2.rewrite_region(filtered.filter)
    printer.print_op(filtered)
    return DeadConstantElim, walker2


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        And now, the transformed filter operation is equivalent to:
        ```
        filter(%table, (lambda t: t.a > 10))
        ```
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        ### Summary

        So, we have now defined a mini-dialect using many of xDSL's features and concepts introduced earlier, including:

        - Operations: we modelled the SQL dialect using `select` and `filter` operations
        - Attributes: we used attributes to encode compile-time available information, from types (the `Bag` type), to constant expressions (e.g. the "cmpi" attribute within `arith.Cmpi`)
        - Mixing dialects: we utilised features from other dialects (`arith`, `scf`, `builtin`)
        - Regions and blocks: we used the operation's _regions_ to model the lambda within `sql.filter` since regions allow nesting of code.
            - The function arguments are modelled as the _block arguments_ of the first block in the region, and the function's code/computation is represented as the block's list of operations.
            - Not all operations need a region, though. For instance `sql.select` does not have a region.
        - xDSL type-checking: xDSL is able to use the typing information within the dialect definition to validate IRs (through `.verify()` calls)
        - Rewrites: we defined two rewrites to transform the xDSL IR, simplifying and optimising the code (`ConstantFolding` and `DeadConstantElim`)
        """
    )
    return


if __name__ == "__main__":
    app.run()



import pytest

from xdsl.context import Context
from xdsl.irdl import (
    IRDLOperation,
    irdl_op_definition,
    var_operand_def,
    var_result_def,
)
from xdsl.parser import Parser
from xdsl.utils.exceptions import ParseError


@irdl_op_definition
class UnknownOp(IRDLOperation):
    name = "test.unknown"
    ops = var_operand_def()
    res = var_result_def()


def check_error(prog: str, line: int, column: int, message: str):
    ctx = Context()
    ctx.load_op(UnknownOp)

    parser = Parser(ctx, prog)
    with pytest.raises(ParseError, match=message) as e:
        parser.parse_operation()

    assert e.value.span.get_line_col() == (line, column)


def test_parser_missing_equal():
    """Test a missing equal sign error."""
    ctx = Context()
    ctx.load_op(UnknownOp)

    prog = """
"test.unknown"() ({
  %0 "test.unknown"() : () -> !i32
}) : () -> ()
"""
    check_error(prog, 3, 5, "Expected '=' after operation result list")


def test_parser_redefined_value():
    """Test an SSA value redefinition error."""
    ctx = Context()
    ctx.load_op(UnknownOp)

    prog = """
"test.unknown"() ({
  %val = "test.unknown"() : () -> i32
  %val = "test.unknown"() : () -> i32
}) : () -> ()
"""
    check_error(prog, 4, 2, "SSA value %val is already defined")


def test_parser_missing_operation_name():
    """Test a missing operation name error."""
    ctx = Context()
    ctx.load_op(UnknownOp)

    prog = """
"test.unknown"() ({
  %val =
}) : () -> ()
"""
    check_error(prog, 4, 0, "operation name expected")



from typing import cast

from xdsl.dialects.builtin import (
    DenseIntOrFPElementsAttr,
)
from xdsl.ir import OpResult
from xdsl.pattern_rewriter import (
    PatternRewriter,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.utils.hints import isa

from ..dialects.toy import ConstantOp, ReshapeOp, TensorTypeF64, TransposeOp


class SimplifyRedundantTranspose(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: TransposeOp, rewriter: PatternRewriter):
        """
        Fold transpose(transpose(x)) -> x
        """
        # Look at the input of the current transpose.
        transpose_input = op.arg
        if not isinstance(transpose_input, OpResult):
            # Input was not produced by an operation, could be a function argument
            return

        transpose_input_op = transpose_input.op
        if not isinstance(transpose_input_op, TransposeOp):
            # Input defined by another transpose? If not, no match.
            return

        rewriter.replace_op(op, [], [transpose_input_op.arg])


class ReshapeReshapeOpPattern(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: ReshapeOp, rewriter: PatternRewriter):
        """
        Reshape(Reshape(x)) = Reshape(x)
        """
        # Look at the input of the current reshape.
        reshape_input = op.arg
        if not isinstance(reshape_input, OpResult):
            # Input was not produced by an operation, could be a function argument
            return

        reshape_input_op = reshape_input.op
        if not isinstance(reshape_input_op, ReshapeOp):
            # Input defined by another transpose? If not, no match.
            return

        t = cast(TensorTypeF64, op.res.type)
        new_op = ReshapeOp.from_input_and_type(reshape_input_op.arg, t)
        rewriter.replace_matched_op(new_op)


class FoldConstantReshapeOpPattern(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: ReshapeOp, rewriter: PatternRewriter):
        """
        Reshaping a constant can be done at compile time
        """
        # Look at the input of the current reshape.
        reshape_input = op.arg
        if not isinstance(reshape_input, OpResult):
            # Input was not produced by an operation, could be a function argument
            return

        reshape_input_op = reshape_input.op
        if not isinstance(reshape_input_op, ConstantOp):
            # Input defined by another transpose? If not, no match.
            return

        assert isa(op.res.type, TensorTypeF64)

        new_value = DenseIntOrFPElementsAttr.create_dense_float(
            type=op.res.type, data=reshape_input_op.value.get_values()
        )
        new_op = ConstantOp(new_value)
        rewriter.replace_matched_op(new_op)



from __future__ import annotations

from abc import ABC

from xdsl.backend.register_type import RegisterType
from xdsl.irdl import (
    irdl_attr_definition,
)


class X86RegisterType(RegisterType, ABC):
    """
    The abstract class for all x86 register types.
    """


X86_INDEX_BY_NAME = {
    "rax": 0,
    "rcx": 1,
    "rdx": 2,
    "rbx": 3,
    "rsp": 4,
    "rbp": 5,
    "rsi": 6,
    "rdi": 7,
    "eax": 0,
    "ecx": 1,
    "edx": 2,
    "ebx": 3,
    "esp": 4,
    "ebp": 5,
    "esi": 6,
    "edi": 7,
    "r8": 8,
    "r9": 9,
    "r10": 10,
    "r11": 11,
    "r12": 12,
    "r13": 13,
    "r14": 14,
    "r15": 15,
}
"""
Mapping of x86 register names to their indices.

See external [documentation](https://wiki.osdev.org/X86-64_Instruction_Encoding#Registers).
"""


@irdl_attr_definition
class GeneralRegisterType(X86RegisterType):
    """
    A scalar x86 register type representing general registers.
    """

    name = "x86.reg"

    @classmethod
    def instruction_set_name(cls) -> str:
        return "x86"

    @classmethod
    def index_by_name(cls) -> dict[str, int]:
        return X86_INDEX_BY_NAME

    @classmethod
    def infinite_register_prefix(cls):
        return "inf_reg_"


UNALLOCATED_GENERAL = GeneralRegisterType.unallocated()
RAX = GeneralRegisterType.from_name("rax")
RCX = GeneralRegisterType.from_name("rcx")
RDX = GeneralRegisterType.from_name("rdx")
RBX = GeneralRegisterType.from_name("rbx")
RSP = GeneralRegisterType.from_name("rsp")
RBP = GeneralRegisterType.from_name("rbp")
RSI = GeneralRegisterType.from_name("rsi")
RDI = GeneralRegisterType.from_name("rdi")

EAX = GeneralRegisterType.from_name("eax")
ECX = GeneralRegisterType.from_name("ecx")
EDX = GeneralRegisterType.from_name("edx")
EBX = GeneralRegisterType.from_name("ebx")
ESP = GeneralRegisterType.from_name("esp")
EBP = GeneralRegisterType.from_name("ebp")
ESI = GeneralRegisterType.from_name("esi")
EDI = GeneralRegisterType.from_name("edi")

R8 = GeneralRegisterType.from_name("r8")
R9 = GeneralRegisterType.from_name("r9")
R10 = GeneralRegisterType.from_name("r10")
R11 = GeneralRegisterType.from_name("r11")
R12 = GeneralRegisterType.from_name("r12")
R13 = GeneralRegisterType.from_name("r13")
R14 = GeneralRegisterType.from_name("r14")
R15 = GeneralRegisterType.from_name("r15")

RFLAGS_INDEX_BY_NAME = {
    "rflags": 0,
}


@irdl_attr_definition
class RFLAGSRegisterType(X86RegisterType):
    """
    A scalar x86 register type representing the RFLAGS register.
    """

    name = "x86.rflags"

    @classmethod
    def instruction_set_name(cls) -> str:
        return "x86"

    @classmethod
    def index_by_name(cls) -> dict[str, int]:
        return RFLAGS_INDEX_BY_NAME

    @classmethod
    def infinite_register_prefix(cls):
        return "inf_rflags_"


UNALLOCATED_RFLAGS = RFLAGSRegisterType.unallocated()
RFLAGS = RFLAGSRegisterType.from_name("rflags")


class X86VectorRegisterType(X86RegisterType):
    """
    The abstract class for all x86 vector register types.
    """


@irdl_attr_definition
class SSERegisterType(X86VectorRegisterType):
    """
    An x86 register type for SSE instructions.
    """

    name = "x86.ssereg"

    @classmethod
    def instruction_set_name(cls) -> str:
        return "SSE"

    @classmethod
    def index_by_name(cls) -> dict[str, int]:
        return SSE_INDEX_BY_NAME

    @classmethod
    def infinite_register_prefix(cls):
        return "inf_sse_"


SSE_INDEX_BY_NAME = {
    "xmm0": 0,
    "xmm1": 1,
    "xmm2": 2,
    "xmm3": 3,
    "xmm4": 4,
    "xmm5": 5,
    "xmm6": 6,
    "xmm7": 7,
    "xmm8": 8,
    "xmm9": 9,
    "xmm10": 10,
    "xmm11": 11,
    "xmm12": 12,
    "xmm13": 13,
    "xmm14": 14,
    "xmm15": 15,
}
"""
Mapping of SSE register names to their indices.

See external # [documentation](https://wiki.osdev.org/X86-64_Instruction_Encoding#Registers).
"""

UNALLOCATED_SSE = SSERegisterType.unallocated()
XMM0 = SSERegisterType.from_name("xmm0")
XMM1 = SSERegisterType.from_name("xmm1")
XMM2 = SSERegisterType.from_name("xmm2")
XMM3 = SSERegisterType.from_name("xmm3")
XMM4 = SSERegisterType.from_name("xmm4")
XMM5 = SSERegisterType.from_name("xmm5")
XMM6 = SSERegisterType.from_name("xmm6")
XMM7 = SSERegisterType.from_name("xmm7")
XMM8 = SSERegisterType.from_name("xmm8")
XMM9 = SSERegisterType.from_name("xmm9")
XMM10 = SSERegisterType.from_name("xmm10")
XMM11 = SSERegisterType.from_name("xmm11")
XMM12 = SSERegisterType.from_name("xmm12")
XMM13 = SSERegisterType.from_name("xmm13")
XMM14 = SSERegisterType.from_name("xmm14")
XMM15 = SSERegisterType.from_name("xmm15")


@irdl_attr_definition
class AVX2RegisterType(X86VectorRegisterType):
    """
    An x86 register type for AVX2 instructions.
    """

    name = "x86.avx2reg"

    @classmethod
    def instruction_set_name(cls) -> str:
        return "AVX2"

    @classmethod
    def index_by_name(cls) -> dict[str, int]:
        return AVX2_INDEX_BY_NAME

    @classmethod
    def infinite_register_prefix(cls):
        return "inf_avx2_"


AVX2_INDEX_BY_NAME = {
    "ymm0": 0,
    "ymm1": 1,
    "ymm2": 2,
    "ymm3": 3,
    "ymm4": 4,
    "ymm5": 5,
    "ymm6": 6,
    "ymm7": 7,
    "ymm8": 8,
    "ymm9": 9,
    "ymm10": 10,
    "ymm11": 11,
    "ymm12": 12,
    "ymm13": 13,
    "ymm14": 14,
    "ymm15": 15,
}
"""
Mapping of AVX2 register names to their indices.

See external # [documentation](https://wiki.osdev.org/X86-64_Instruction_Encoding#Registers).
"""

UNALLOCATED_AVX2 = AVX2RegisterType.unallocated()
YMM0 = AVX2RegisterType.from_name("ymm0")
YMM1 = AVX2RegisterType.from_name("ymm1")
YMM2 = AVX2RegisterType.from_name("ymm2")
YMM3 = AVX2RegisterType.from_name("ymm3")
YMM4 = AVX2RegisterType.from_name("ymm4")
YMM5 = AVX2RegisterType.from_name("ymm5")
YMM6 = AVX2RegisterType.from_name("ymm6")
YMM7 = AVX2RegisterType.from_name("ymm7")
YMM8 = AVX2RegisterType.from_name("ymm8")
YMM9 = AVX2RegisterType.from_name("ymm9")
YMM10 = AVX2RegisterType.from_name("ymm10")
YMM11 = AVX2RegisterType.from_name("ymm11")
YMM12 = AVX2RegisterType.from_name("ymm12")
YMM13 = AVX2RegisterType.from_name("ymm13")
YMM14 = AVX2RegisterType.from_name("ymm14")
YMM15 = AVX2RegisterType.from_name("ymm15")


@irdl_attr_definition
class AVX512RegisterType(X86VectorRegisterType):
    """
    An x86 register type for AVX512 instructions.
    """

    name = "x86.avx512reg"

    @classmethod
    def instruction_set_name(cls) -> str:
        return "AVX512"

    @classmethod
    def index_by_name(cls) -> dict[str, int]:
        return X86AVX512_INDEX_BY_NAME

    @classmethod
    def infinite_register_prefix(cls):
        return "inf_avx512_"


X86AVX512_INDEX_BY_NAME = {
    "zmm0": 0,
    "zmm1": 1,
    "zmm2": 2,
    "zmm3": 3,
    "zmm4": 4,
    "zmm5": 5,
    "zmm6": 6,
    "zmm7": 7,
    "zmm8": 8,
    "zmm9": 9,
    "zmm10": 10,
    "zmm11": 11,
    "zmm12": 12,
    "zmm13": 13,
    "zmm14": 14,
    "zmm15": 15,
    "zmm16": 16,
    "zmm17": 17,
    "zmm18": 18,
    "zmm19": 19,
    "zmm20": 20,
    "zmm21": 21,
    "zmm22": 22,
    "zmm23": 23,
    "zmm24": 24,
    "zmm25": 25,
    "zmm26": 26,
    "zmm27": 27,
    "zmm28": 28,
    "zmm29": 29,
    "zmm30": 30,
    "zmm31": 31,
}
"""
Mapping of AVX512 register names to their indices.

See external # [documentation](https://wiki.osdev.org/X86-64_Instruction_Encoding#Registers).
"""

UNALLOCATED_AVX512 = AVX512RegisterType.unallocated()
ZMM0 = AVX512RegisterType.from_name("zmm0")
ZMM1 = AVX512RegisterType.from_name("zmm1")
ZMM2 = AVX512RegisterType.from_name("zmm2")
ZMM3 = AVX512RegisterType.from_name("zmm3")
ZMM4 = AVX512RegisterType.from_name("zmm4")
ZMM5 = AVX512RegisterType.from_name("zmm5")
ZMM6 = AVX512RegisterType.from_name("zmm6")
ZMM7 = AVX512RegisterType.from_name("zmm7")
ZMM8 = AVX512RegisterType.from_name("zmm8")
ZMM9 = AVX512RegisterType.from_name("zmm9")
ZMM10 = AVX512RegisterType.from_name("zmm10")
ZMM11 = AVX512RegisterType.from_name("zmm11")
ZMM12 = AVX512RegisterType.from_name("zmm12")
ZMM13 = AVX512RegisterType.from_name("zmm13")
ZMM14 = AVX512RegisterType.from_name("zmm14")
ZMM15 = AVX512RegisterType.from_name("zmm15")
ZMM16 = AVX512RegisterType.from_name("zmm16")
ZMM17 = AVX512RegisterType.from_name("zmm17")
ZMM18 = AVX512RegisterType.from_name("zmm18")
ZMM19 = AVX512RegisterType.from_name("zmm19")
ZMM20 = AVX512RegisterType.from_name("zmm20")
ZMM21 = AVX512RegisterType.from_name("zmm21")
ZMM22 = AVX512RegisterType.from_name("zmm22")
ZMM23 = AVX512RegisterType.from_name("zmm23")
ZMM24 = AVX512RegisterType.from_name("zmm24")
ZMM25 = AVX512RegisterType.from_name("zmm25")
ZMM26 = AVX512RegisterType.from_name("zmm26")
ZMM27 = AVX512RegisterType.from_name("zmm27")
ZMM28 = AVX512RegisterType.from_name("zmm28")
ZMM29 = AVX512RegisterType.from_name("zmm29")
ZMM30 = AVX512RegisterType.from_name("zmm30")
ZMM31 = AVX512RegisterType.from_name("zmm31")



from xdsl.builder import ImplicitBuilder
from xdsl.dialects import arith, func, memref
from xdsl.dialects.builtin import (
    DenseIntOrFPElementsAttr,
    IndexType,
    ModuleOp,
    StringAttr,
    TensorType,
    i32,
)
from xdsl.interpreter import Interpreter
from xdsl.interpreters.arith import ArithFunctions
from xdsl.interpreters.memref import MemRefFunctions
from xdsl.interpreters.shaped_array import ShapedArray
from xdsl.interpreters.utils.ptr import TypedPtr

interpreter = Interpreter(ModuleOp([]), index_bitwidth=32)
interpreter.register_implementations(ArithFunctions())
interpreter.register_implementations(MemRefFunctions())

index = IndexType()


def test_functions():
    alloc_op = memref.AllocOp.get(i32, None, (2, 3))
    zero_op = arith.ConstantOp.from_int_and_width(0, index)
    one_op = arith.ConstantOp.from_int_and_width(1, index)
    forty_two_op = arith.ConstantOp.from_int_and_width(42, 32)
    store_op = memref.StoreOp.get(forty_two_op, alloc_op, (zero_op, one_op))
    load_42_op = memref.LoadOp.get(alloc_op, (zero_op, one_op))
    dealloc_op = memref.DeallocOp.get(alloc_op)

    (shaped_array,) = interpreter.run_op(alloc_op, ())

    assert shaped_array == ShapedArray(
        TypedPtr.new_index((0,) * 6, interpreter.index_bitwidth), [2, 3]
    )
    (zero,) = interpreter.run_op(zero_op, ())
    (one,) = interpreter.run_op(one_op, ())

    (forty_two_0,) = interpreter.run_op(forty_two_op, ())
    store_res = interpreter.run_op(store_op, (forty_two_0, shaped_array, zero, one))
    assert store_res == ()
    (forty_two_1,) = interpreter.run_op(load_42_op, (shaped_array, zero, one))
    assert forty_two_1 == 42

    dealloc_res = interpreter.run_op(dealloc_op, (shaped_array,))
    assert dealloc_res == ()


def test_memref_get_global():
    memref_type = memref.MemRefType(i32, (2, 2))
    tensor_type = TensorType(i32, (2, 2))
    module = ModuleOp([])
    with ImplicitBuilder(module.body):
        memref.GlobalOp.get(
            StringAttr("my_global"),
            memref_type,
            DenseIntOrFPElementsAttr.from_list(tensor_type, [1, 2, 3, 4]),
            sym_visibility=StringAttr("public"),
        )
        with ImplicitBuilder(func.FuncOp("main", ((), ())).body):
            fetch = memref.GetGlobalOp("my_global", memref_type)

    interpreter = Interpreter(module, index_bitwidth=32)
    interpreter.register_implementations(MemRefFunctions())

    (result,) = interpreter.run_op(fetch, ())
    assert result == ShapedArray(
        TypedPtr.new_index([1, 2, 3, 4], index_bitwidth=interpreter.index_bitwidth),
        [2, 2],
    )



from xdsl.dialects import eqsat, pdl
from xdsl.dialects.builtin import (
    IntegerType,
    StringAttr,
)
from xdsl.interpreters.eqsat_pdl import EqsatPDLMatcher


def test_match_type():
    matcher = EqsatPDLMatcher()

    pdl_op = pdl.TypeOp()
    ssa_value = pdl_op.result
    ssa_value = eqsat.EClassOp(ssa_value).result
    xdsl_value = StringAttr("a")

    # New value
    assert matcher.match_type(ssa_value, pdl_op, xdsl_value)
    assert matcher.matching_context == {ssa_value: xdsl_value}

    # Same value
    assert matcher.match_type(ssa_value, pdl_op, xdsl_value)
    assert matcher.matching_context == {ssa_value: xdsl_value}

    # Other value
    assert not matcher.match_type(ssa_value, pdl_op, StringAttr("b"))
    assert matcher.matching_context == {ssa_value: xdsl_value}


def test_match_fixed_type():
    matcher = EqsatPDLMatcher()

    pdl_op = pdl.TypeOp(IntegerType(32))
    xdsl_value = IntegerType(32)
    ssa_value = pdl_op.result
    ssa_value = eqsat.EClassOp(ssa_value).result

    assert matcher.match_type(ssa_value, pdl_op, xdsl_value)
    assert matcher.matching_context == {ssa_value: xdsl_value}


def test_not_match_fixed_type():
    matcher = EqsatPDLMatcher()

    pdl_op = pdl.TypeOp(IntegerType(64))
    xdsl_value = IntegerType(32)
    ssa_value = pdl_op.result
    ssa_value = eqsat.EClassOp(ssa_value).result

    assert not matcher.match_type(ssa_value, pdl_op, xdsl_value)
    assert matcher.matching_context == {}



from xdsl.context import Context
from xdsl.dialects.builtin import (
    Builtin,
    StringAttr,
)
from xdsl.dialects.test import Test, TestOp
from xdsl.interactive.passes import AvailablePass
from xdsl.interactive.rewrites import get_all_possible_rewrites
from xdsl.parser import Parser
from xdsl.pattern_rewriter import (
    PatternRewriter,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.transforms.individual_rewrite import ApplyIndividualRewritePass


class Rewrite(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: TestOp, rewriter: PatternRewriter):
        if op.attributes["label"] != StringAttr("a"):
            return
        rewriter.replace_matched_op(TestOp(attributes={"label": StringAttr("c")}))


def test_get_all_possible_rewrite():
    # build module
    prog = """
    builtin.module {
    "test.op"() {"label" = "a"} : () -> ()
    "test.op"() {"label" = "a"} : () -> ()
    "test.op"() {"label" = "b"} : () -> ()
    }
    """

    ctx = Context()
    ctx.load_dialect(Builtin)
    ctx.load_dialect(Test)

    parser = Parser(ctx, prog)
    module = parser.parse_module()

    expected_res = [
        AvailablePass(
            display_name='TestOp("test.op"() {label = "a"} : () -> ()):test.op:TestRewrite',
            module_pass=ApplyIndividualRewritePass(1, "test.op", "TestRewrite"),
        ),
        AvailablePass(
            display_name='TestOp("test.op"() {label = "a"} : () -> ()):test.op:TestRewrite',
            module_pass=ApplyIndividualRewritePass(2, "test.op", "TestRewrite"),
        ),
    ]

    res = get_all_possible_rewrites(module, {"test.op": {"TestRewrite": Rewrite()}})
    assert res == expected_res



"""
A dialect that represents at the highest level of abstraction the capabilities of the
[Snitch](https://github.com/pulp-platform/snitch_cluster) accelerator core, as used in
[Occamy](https://github.com/pulp-platform/occamy) and others.

The core aims to optimise for performance per watt, by replacing caches and branch
prediction logic with streaming registers and fixed-repetition loops. This dialect models
the streaming functionality of the Snitch core.

`snitch_stream.stride_pattern_type` represents a specification of the order in which
elements of a streamed region of memory will be read from or written to.

`snitch_stream.stride_pattern` creates a value storing the above specification.

`snitch_stream.streaming_region` encapsulates a region of code where the streams are
valid. According to the Snitch ABI, within this region, the registers `ft0` to `ftn`,
where `n` is the number of streaming registers, have a restricted functionality. If the
register is configured as a readable stream register, then it cannot be written to, and
if the register is configured as a writable stream register, then it cannot be read from.
"""

from __future__ import annotations

from collections.abc import Iterator, Sequence
from itertools import product
from typing import cast

from xdsl.dialects import riscv
from xdsl.dialects.builtin import (
    ArrayAttr,
    IntAttr,
)
from xdsl.ir import (
    Attribute,
    Dialect,
    ParametrizedAttribute,
    Region,
    SSAValue,
)
from xdsl.irdl import (
    AttrSizedOperandSegments,
    IRDLOperation,
    ParameterDef,
    irdl_attr_definition,
    irdl_op_definition,
    prop_def,
    region_def,
    traits_def,
    var_operand_def,
)
from xdsl.parser import AttrParser, Parser
from xdsl.printer import Printer
from xdsl.traits import NoTerminator
from xdsl.utils.exceptions import VerifyException


@irdl_attr_definition
class StridePattern(ParametrizedAttribute):
    """
    Attribute representing the order and offsets in which elements will be read from or
    written to a stream.

    ```
    // 2D access pattern
    #pat = #snitch_stream.stride_pattern<ub = [16, 8], strides = [8, 128]>
    // Corresponds to the following locations
    // for i in range(16):
    //   for j in range(8):
    //     yield i * 8 + j * 128
    // Note that the upper bounds and strides go from the outermost loop inwards
    ```
    """

    name = "snitch_stream.stride_pattern"

    ub: ParameterDef[ArrayAttr[IntAttr]]
    strides: ParameterDef[ArrayAttr[IntAttr]]
    repeat: ParameterDef[IntAttr]
    """
    Number of times an element will be repeated when loaded, default is 1.
    """

    def __init__(
        self,
        ub: ArrayAttr[IntAttr],
        strides: ArrayAttr[IntAttr],
        repeat: IntAttr = IntAttr(1),
    ):
        super().__init__((ub, strides, repeat))

    @classmethod
    def parse_parameters(cls, parser: AttrParser) -> Sequence[Attribute]:
        with parser.in_angle_brackets():
            parser.parse_identifier("ub")
            parser.parse_punctuation("=")
            ub = ArrayAttr(
                IntAttr(i)
                for i in parser.parse_comma_separated_list(
                    parser.Delimiter.SQUARE, parser.parse_integer
                )
            )
            parser.parse_punctuation(",")
            parser.parse_identifier("strides")
            parser.parse_punctuation("=")
            strides = ArrayAttr(
                IntAttr(i)
                for i in parser.parse_comma_separated_list(
                    parser.Delimiter.SQUARE, parser.parse_integer
                )
            )
            if parser.parse_optional_punctuation(","):
                parser.parse_identifier("repeat")
                parser.parse_punctuation("=")
                repeat = parser.parse_integer(allow_boolean=False, allow_negative=False)
            else:
                repeat = 1
            return (ub, strides, IntAttr(repeat))

    def print_parameters(self, printer: Printer) -> None:
        with printer.in_angle_brackets():
            printer.print_string("ub = [")
            printer.print_list(self.ub, lambda attr: printer.print(attr.data))
            printer.print_string("], strides = [")
            printer.print_list(self.strides, lambda attr: printer.print(attr.data))
            printer.print_string("]")
            if self.repeat.data != 1:
                printer.print_string(f", repeat = {self.repeat.data}")

    @staticmethod
    def from_bounds_and_strides(
        ub: Sequence[int], strides: Sequence[int], repeat: int = 1
    ) -> StridePattern:
        return StridePattern(
            ArrayAttr(IntAttr(i) for i in ub),
            ArrayAttr(IntAttr(i) for i in strides),
            IntAttr(repeat),
        )

    def rank(self):
        return len(self.ub)

    def verify(self) -> None:
        if len(self.ub) != len(self.strides):
            raise VerifyException(
                f"Expect stride pattern upper bounds {self.ub} to be equal in length to strides {self.strides}"
            )

    def offset_iter(self) -> Iterator[int]:
        for indices in product(*(range(bound.data) for bound in self.ub.data)):
            indices: tuple[int, ...] = indices
            offset = sum(
                index * stride.data
                for (index, stride) in zip(indices, self.strides.data)
            )
            for _ in range(self.repeat.data):
                yield offset

    def offsets(self) -> tuple[int, ...]:
        return tuple(self.offset_iter())

    def simplified(self) -> StridePattern:
        """
        Return a stride pattern that specifies the same iteration space, but with folded
        perfectly nested outermost loops.

        e.g.

        ```
        stride_pattern<ub = [2, 3, 4], strides = [12, 4, 1]>
        ->
        stride_pattern<ub = [24], strides = [1]
        ```
        """
        if len(self.ub) < 2:
            return self

        tuples = tuple(
            (bound.data, stride.data)
            for bound, stride in zip(self.ub.data, self.strides.data)
            # Exclude single iteration bounds
            if bound.data != 1
        )

        if not tuples:
            # All bounds are 1
            return StridePattern.from_bounds_and_strides(
                (1,), (self.strides.data[-1].data,)
            )

        # Outermost bound and stride
        ub0, s0 = tuples[0]

        # Start with the second outermost loop bounds
        second_outermost_dim = 1
        while second_outermost_dim < len(tuples):
            # Next bound and stride to fold into outermost
            ubd, sd = tuples[second_outermost_dim]
            if s0 == ubd * sd:
                # The second outermost loop is perfectly nested in outermost
                ub0 = ub0 * ubd
                s0 = sd
                # Decrement the index into tuples for what the new second outermost loop
                # bound is
                second_outermost_dim += 1
            else:
                # The second outermost loop does not match, do not try to further simplify
                break

        # ub and s include the new outermost bound and stride,
        # followed by all the tuples up to and including the second outermost dim
        ub = (ub0, *(bound for bound, _ in tuples[second_outermost_dim:]))
        s = (s0, *(stride for _, stride in tuples[second_outermost_dim:]))

        if s[-1] == 0:
            repeat = ub[-1] * self.repeat.data
            ub = ub[:-1]
            s = s[:-1]
        else:
            repeat = self.repeat.data

        return StridePattern.from_bounds_and_strides(ub, s, repeat)


@irdl_op_definition
class StreamingRegionOp(IRDLOperation):
    """
    An operation that creates streams from access patterns, which are only available to
    read from and write to within the body of the operation.

    According to the Snitch ABI, within this region, the registers `ft0` to `ftn`,
    where `n` is the number of streaming registers, have a restricted functionality. If the
    register is configured as a readable stream register, then it cannot be written to, and
    if the register is configured as a writable stream register, then it cannot be read from.
    """

    name = "snitch_stream.streaming_region"

    inputs = var_operand_def(riscv.IntRegisterType)
    """
    Pointers to memory buffers that will be streamed. The corresponding stride pattern
    defines the order in which the elements of the input buffers will be read.
    """
    outputs = var_operand_def(riscv.IntRegisterType)
    """
    Pointers to memory buffers that will be streamed. The corresponding stride pattern
    defines the order in which the elements of the input buffers will be written to.
    """
    stride_patterns = prop_def(ArrayAttr[StridePattern])
    """
    Stride patterns that define the order of the input and output streams. If there is
    one stride pattern, and more inputs and outputs, the stride pattern is applied to all
    the streams.
    """

    body = region_def("single_block")

    irdl_options = [AttrSizedOperandSegments(as_property=True)]

    traits = traits_def(NoTerminator())

    def __init__(
        self,
        inputs: Sequence[SSAValue],
        outputs: Sequence[SSAValue],
        stride_patterns: ArrayAttr[StridePattern],
        body: Region,
    ) -> None:
        super().__init__(
            operands=[inputs, outputs],
            regions=[body],
            properties={
                "stride_patterns": stride_patterns,
            },
        )

    def print(self, printer: Printer):
        with printer.indented():
            printer.print_string(" {")
            if self.stride_patterns.data:
                printer.print_string("\npatterns = [")
                with printer.indented():
                    printer.print_list(
                        self.stride_patterns.data,
                        lambda attr: printer.print("\n", attr),
                        delimiter=",",
                    )
                printer.print_string("\n]")
            else:
                printer.print_string("\npatterns = []")
        printer.print_string("\n}")

        if self.inputs:
            printer.print_string(" ins(")
            printer.print_list(self.inputs, printer.print_ssa_value)
            printer.print_string(" : ")
            printer.print_list(self.inputs.types, printer.print_attribute)
            printer.print_string(")")

        if self.outputs:
            printer.print_string(" outs(")
            printer.print_list(self.outputs, printer.print_ssa_value)
            printer.print_string(" : ")
            printer.print_list(self.outputs.types, printer.print_attribute)
            printer.print_string(")")

        if self.attributes:
            printer.print(" attributes = ")
            printer.print_op_attributes(self.attributes)

        printer.print_string(" ")
        printer.print_region(self.body)

    @classmethod
    def parse(cls, parser: Parser) -> StreamingRegionOp:
        parser.parse_punctuation("{")
        parser.parse_identifier("stride_patterns")
        parser.parse_punctuation("=")

        patterns = parser.parse_attribute()
        if not isinstance(patterns, ArrayAttr):
            parser.raise_error(f"Expected ArrayAttr {patterns}")
        patterns = cast(ArrayAttr[Attribute], patterns)
        for pattern in patterns:
            if not isinstance(pattern, StridePattern):
                parser.raise_error(f"Expected StridePattern {pattern}")
        patterns = cast(ArrayAttr[StridePattern], patterns)

        parser.parse_punctuation("}")

        pos = parser.pos
        if parser.parse_optional_characters("ins"):
            parser.parse_punctuation("(")
            unresolved_ins = parser.parse_comma_separated_list(
                Parser.Delimiter.NONE, parser.parse_unresolved_operand
            )
            parser.parse_punctuation(":")
            ins_types = parser.parse_comma_separated_list(
                Parser.Delimiter.NONE, parser.parse_type
            )
            parser.parse_punctuation(")")
            ins = parser.resolve_operands(unresolved_ins, ins_types, pos)
        else:
            ins = ()

        pos = parser.pos
        if parser.parse_optional_characters("outs"):
            parser.parse_punctuation("(")
            unresolved_outs = parser.parse_comma_separated_list(
                Parser.Delimiter.NONE, parser.parse_unresolved_operand
            )
            parser.parse_punctuation(":")
            outs_types = parser.parse_comma_separated_list(
                Parser.Delimiter.NONE, parser.parse_type
            )
            parser.parse_punctuation(")")
            outs = parser.resolve_operands(unresolved_outs, outs_types, pos)
        else:
            outs = ()

        if parser.parse_optional_keyword("attributes"):
            parser.parse_punctuation("=")
            extra_attrs = parser.expect(
                parser.parse_optional_attr_dict, "expect extra attributes"
            )
        else:
            extra_attrs = {}

        body = parser.parse_region()

        generic = cls(
            ins,
            outs,
            patterns,
            body,
        )
        generic.attributes |= extra_attrs

        return generic


SnitchStream = Dialect(
    "snitch_stream",
    [
        StreamingRegionOp,
    ],
    [
        StridePattern,
    ],
)
